"""Universal interface for wrapper developers."""
from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from base64 import b64encode
from collections.abc import Collection
from datetime import datetime, timedelta
from enum import IntEnum
from itertools import chain
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import (
    TYPE_CHECKING,
    BinaryIO,
    Dict,
    Iterable,
    Iterator,
    Optional,
    Set,
    Tuple,
    Union,
)

from rdflib import RDF, BNode, Graph, URIRef
from rdflib.graph import ModificationException, ReadOnlyGraphAggregate
from rdflib.plugins.stores.memory import SimpleMemory
from rdflib.query import Result
from rdflib.store import Store
from rdflib.term import Node

from simphony_osp.session.session import Session
from simphony_osp.utils.datatypes import Pattern, Triple
from simphony_osp.utils.other import take
from simphony_osp.utils.simphony_namespace import simphony_namespace

if TYPE_CHECKING:
    from simphony_osp.ontology.entity import OntologyEntity

__all__ = [
    "BufferType",
    "BufferedSimpleMemoryStore",
    "Interface",
    "InterfaceDriver",
]

logger = logging.getLogger(__name__)


class BufferType(IntEnum):
    """Enum representing the two possible types of triple buffers.

    - ADDED: For triples that have been added.
    - DELETED: For triples that have been deleted.
    """

    ADDED = 0
    DELETED = 1


class BufferedSimpleMemoryStore(SimpleMemory):
    """A transaction-aware extension of the RDFLib SimpleMemory store."""

    _buffers: Dict[BufferType, Graph]
    """Holds added and removed triples until a commit is performed."""

    # RDFLib
    # ↓ -- ↓

    transaction_aware = True

    def __init__(self, *args, **kwargs):
        """Initialize the store."""
        self._reset_buffers()  # Creates the buffers for the first time.
        super().__init__(*args, **kwargs)

    def open(self, *args, **kwargs) -> Optional[int]:
        """Opens the store.

        Resets the buffers before performing the operation.
        """
        self._reset_buffers()
        return super().open(*args, **kwargs)

    def close(self, *args, **kwargs) -> None:
        """Closes the store.

        Resets the buffers after having closed the `SimpleMemory` store.
        """
        result = super().close(*args, **kwargs)
        self._reset_buffers()
        return result

    def add(self, triple: Triple, context: Graph, quoted=False) -> None:
        """Add triples to the store.

        Since this store is meant to be transaction-aware, the changes are
        just buffered until a commit is performed.
        """
        self._buffers[BufferType.DELETED].remove(triple)
        self._buffers[BufferType.ADDED].add(triple)

    def remove(
        self, triple_pattern: Pattern, context: Optional[Graph] = None
    ) -> None:
        """Remove triples from the store.

        Since this store is transaction-aware, the changes are buffered
        until a commit is performed.
        """
        self._buffers[BufferType.ADDED].remove(triple_pattern)
        existing_triples = (
            (s, p, o) for (s, p, o), ctx in self.triples(triple_pattern)
        )
        for triple in existing_triples:
            self._buffers[BufferType.DELETED].add(triple)

    def triples(
        self, triple_pattern: Pattern, context=None, ignore_buffers=False
    ) -> Iterator[Tuple[Triple, Iterator[Graph]]]:
        """Fetch triples from the store that match the given pattern.

        The triples from the underlying `SimpleMemory` store are combined
        with the contents of the buffers to produce the final result
        """
        triple_pool = filter(
            lambda x: x
            not in chain(
                self._buffers[BufferType.ADDED].triples(triple_pattern),
                self._buffers[BufferType.DELETED].triples(triple_pattern),
            ),
            (
                (s, p, o)
                for (s, p, o), ctx in super().triples(triple_pattern, None)
            ),
        )
        # Include added triples (previously they were excluded in order not
        # to duplicate triples).
        triple_pool = chain(
            triple_pool,
            self._buffers[BufferType.ADDED].triples(triple_pattern),
        )

        yield from ((triple, iter(())) for triple in triple_pool)

    def query(self, *args, **kwargs) -> Result:
        """Perform a SPARQL query on the store."""
        # TODO: use the store's implementation when the buffers are empty.
        raise NotImplementedError  # Just use RDFLib's query processor.

    def update(self, *args, **kwargs) -> None:
        """Perform s SPARQL update query on the store."""
        # TODO: use the store's implementation when the buffers are empty.
        raise NotImplementedError  # Just use RDFLib's query processor.

    def commit(self) -> None:
        """Commit buffered changes to the underlying `SimpleMemory` store."""
        # Remove triples from deleted buffer.
        subclass_method = self.triples
        setattr(self, "triples", super().triples)
        for triple in self._buffers[BufferType.DELETED]:
            super().remove(triple, None)
        setattr(self, "triples", subclass_method)

        # Add triples from added buffer.
        subclass_method = self.add
        setattr(self, "add", super().add)
        super().addN(
            (s, p, o, self._buffers[BufferType.ADDED])
            for s, p, o in self._buffers[BufferType.ADDED]
        )
        setattr(self, "add", subclass_method)

        self._reset_buffers()

        super().commit()

    def rollback(self) -> None:
        """Revert buffered changes."""
        self._reset_buffers()
        super().rollback()

    # RDFLib
    # ↑ -- ↑

    def _reset_buffers(self) -> None:
        """Replaces the existing buffers by empty buffers."""
        self._buffers = {
            buffer_type: Graph(SimpleMemory()) for buffer_type in BufferType
        }


class CacheSizeException(RuntimeError):
    """Raised when  triples do not fit in the cache of `InterfaceDriver`."""


class InterfaceDriver(Store):
    """RDFLib store acting as intermediary between SimPhoNy and wrappers.

    Offers a triplestore interface for SimPhoNy to interact with wrappers.

     SimPhoNy <--> InterfaceDriver <--> Interface

    The store is transaction aware (needs a commit action to save the changes
    to the wrapper). Otherwise, one would have to update an entity
    every time a single triple from the entity is added.
    """

    interface: Interface
    """The interface that the driver controls."""

    _buffer_caught: Dict[BufferType, Graph]
    """Holds triples that should not be added to the interface's base graph."""

    _buffer_uncaught: Dict[BufferType, Graph]
    """Holds triples that should be added to the interface's base graph."""

    _queue: Dict[URIRef, Optional[BinaryIO]]
    """Holds handles of files to be uploaded, URIs of files to be deleted."""

    _ontology: Optional[Session]
    """Ontology to be used on the session's shown to the wrapper developer."""

    _cache: Graph
    """Holds cached triples when the interface's cache is enabled."""

    _cached_patterns: Dict[Pattern, datetime]
    """Holds currently cached patterns and the time they were cached."""

    _cache_size: int = 100000
    """Maximum size of the cache in triples."""

    _cache_time: timedelta = timedelta(seconds=0.3)
    """Maximum time to spend in caching after making a request.

    When a request for triples involving a specific subject is made,
    the caching algorithm asks for all the triples involving the subject.
    This call is not subject to a time limit.

    However, in a second stage, if the interface is merely based on a base
    graph (meaning requests can be made using SPARQL), the algorithm will
    also try to cache the children of the such subject, the parent objects,
    and all of its neighbors. Since these operations can take a lot of extra
    time, this parameter fixes a time limit for the cumulative time it
    takes to perform all of these requests. When such caching operations
    take longer than this time, they are interrupted.
    """

    _cache_step: int = 100
    """Number of triples to fetch at once when filling the cache.

    Since SimPhoNy is single-threaded software, the queries described above
    (on the docstring of `_cache_time`) are blocking. This means that if one
    tries to fetch too many items at once from the interface, it may take
    significantly longer than the `_cache_time` to do so, but the operation
    cannot be canceled until all the items have been fetched.

    That is the reason behind this parameter: if it is small enough, one can
    check how long has the caching algorithm been running at smaller
    intervals. Please note that this assumes that asking the interface to
    yield one triple takes a much smaller amount of time than `_cache_time`.
    This is not necessarily always the case (for example if the underlying
    interface produces triples in chunks).
    """

    _file_cache: Optional[TemporaryDirectory] = None
    """Holds files that are accessed while queued.

    Files are queued as byte streams (which can be exhausted). Therefore,
    if a user needs to access a file before it is committed, the byte stream
    needs to be copied somewhere to be committed later. This also applies
    when files need to be accessed during commit by a wrapper (as they are
    truly sent to the wrapper after the `commit` method has finished).
    """

    # RDFLib
    # ↓ -- ↓

    context_aware: bool = False
    formula_aware: bool = False
    transaction_aware: bool = True
    graph_aware: bool = False

    def __init__(
        self,
        *args,
        interface: Interface,
        ontology: Optional[Session] = None,
        **kwargs,
    ):
        """Initialize the InterfaceDriver.

        The initialization assigns an interface to the store and creates
        buffers. Then the usual RDFLib's store initialization follows.
        """
        if not isinstance(interface, Interface):
            raise ValueError("No valid interface provided.")

        interface.close()
        self.interface = interface

        self._buffer_uncaught = {
            buffer_type: Graph(SimpleMemory()) for buffer_type in BufferType
        }
        self._buffer_caught = {
            buffer_type: Graph(SimpleMemory()) for buffer_type in BufferType
        }

        self._ontology = ontology
        self._queue = dict()
        self._cache = Graph("SimpleMemory")
        self._cached_patterns = dict()
        super().__init__(*args, **kwargs)

    def open(self, configuration: str, create: bool = False) -> None:
        """Asks the interface to open the data source."""
        # Reset buffers
        self._buffer_uncaught = {
            buffer_type: Graph(SimpleMemory()) for buffer_type in BufferType
        }
        self._buffer_caught = {
            buffer_type: Graph(SimpleMemory()) for buffer_type in BufferType
        }

        # Reset cache
        self._cache.remove((None, None, None))

        # Set-up file bytestream cache
        self._file_cache = TemporaryDirectory()

        self.interface.open(configuration, create)

        # The interface can set its base graph when `open` is called. If not
        # set by the interface, then an in-memory transaction-aware graph is
        # used. A session based on this graph is also provided to the
        # interface.
        if self.interface.base is None:
            self.interface.base = Graph(store=BufferedSimpleMemoryStore())
        self.interface.session_base = Session(
            base=self.interface.base, ontology=self._ontology
        )

        # Call the populate method of the interface on base graph/session.
        # The populate method is meant to act on the base graph.
        session = self.interface.session_base
        self.interface.session = session
        try:
            session.lock()
            # Session must not close when exiting its context manager (just
            # in case the user enters the context manager again inside
            # populate).
            with session:
                self.interface.populate()
                session.commit()
        finally:
            session.unlock()
            self.interface.session_base = None
            self.interface.session = None

    def close(self, commit_pending_transaction: bool = False) -> None:
        """Tells the interface to close the data source.

        Args:
            commit_pending_transaction: commits uncommitted changes when
                true before closing the data source.
        """
        if commit_pending_transaction:
            self.commit()
        self.interface.close()

        # Clear the cache
        self._cache.remove((None, None, None))

        # Clear all sessions, graphs, and entities provided to the interface.
        self.interface.base = None
        self.interface.old_graph = None
        self.interface.new_graph = None
        self.interface.buffer = None
        self.interface.session_base = None
        self.interface.session_old = None
        self.interface.session_new = None
        self.interface.added = None
        self.interface.updated = None
        self.interface.deleted = None

        # Clear bytestream cache
        if self._file_cache is not None:
            self._file_cache.cleanup()
            self._file_cache = None

    def add(self, triple: Triple, context: Graph, quoted=False) -> None:
        """Adds triples to the interface.

        Since the actual addition happens during a commit, this method just
        buffers the changes.
        """
        buffer = (
            self._buffer_caught
            if hasattr(self.interface, "add")
            and not self.interface.add(triple)
            else self._buffer_uncaught
        )

        buffer[BufferType.DELETED].remove(triple)
        buffer[BufferType.ADDED].add(triple)

    def remove(
        self, triple_pattern: Pattern, context: Optional[Graph] = None
    ) -> None:
        """Remove triples from the interface.

        Since the actual removal happens during a commit, this method just
        buffers the changes.
        """
        for buffer in (self._buffer_uncaught, self._buffer_caught):
            buffer[BufferType.ADDED].remove(triple_pattern)
            existing_triples = (
                self.interface.base.triples(triple_pattern)
                if buffer is self._buffer_uncaught
                else (
                    set(self.interface.remove(triple_pattern))
                    if hasattr(self.interface, "remove")
                    else set()
                )
            )
            for triple in existing_triples:
                buffer[BufferType.DELETED].add(triple)

    def triples(
        self, triple_pattern: Pattern, context=None, ignore_buffers=False
    ) -> Iterator[Tuple[Triple, Graph]]:
        """Query the interface for triple patterns.

        This method combines the information on the buffers with the one
        provided by the interface. Moreover, if defined, it will use the
        `triples` method of the interface so that it can watch which triples
        are being retrieved or produce triples from its data structures on
        the fly.

        In addition, this method is responsible for trying to obtain the
        triples from the cache if it is enabled, and for filling it on cache
        misses.
        """
        # Determine source of triples.
        if self.interface.cache and self._cached(triple_pattern):
            query_method = self._cache.triples
            fill_cache, fill_cache_sparql = False, False
        elif hasattr(self.interface, "triples"):
            query_method = self.interface.triples
            fill_cache, fill_cache_sparql = True, False
        else:
            query_method = self.interface.base.triples
            fill_cache, fill_cache_sparql = True, True

        # Manage the triple cache.
        if fill_cache and self.interface.cache:
            """Fill the cache using requests based on triple patterns.

            This stage caches either the requested triple pattern, or if the
            triple pattern contains a subject, all the information about the
            subject.
            """
            timestamp = datetime.now()

            s_p, p_p, o_p = triple_pattern
            if s_p is not None:
                """The requested pattern contains a subject."""
                cache_pattern = (s_p, None, None)
            else:
                """No subject in the triple pattern.

                Cache the triple pattern only.
                """
                cache_pattern = triple_pattern
                fill_cache_sparql = False
            all_triples = set(query_method(cache_pattern))

            def query_method(pattern_triples: Pattern) -> Iterator[Triple]:
                """Filter the retrieved triples.

                Yields only the triples matching the given pattern.
                """
                pat_s, pat_p, pat_o = pattern_triples
                yield from (
                    (a, b, c)
                    for a, b, c in all_triples
                    if all(
                        (
                            pat_s is None or a == pat_s,
                            pat_p is None or b == pat_p,
                            pat_o is None or c == pat_o,
                        )
                    )
                )

            try:
                """Fill the cache with the retrieved triples."""
                self._fill(cache_pattern, all_triples)
            except CacheSizeException:
                """No space for the retrieved triples. No action."""
                pass

            if fill_cache_sparql:
                """Fill the cache using requests based on SPARQL.

                This stage allows additional triples related to the children of
                the object, the parent object and the neighbors to be cached.
                """
                base = self.interface.base

                if not isinstance(s_p, BNode):
                    queries = (
                        f"""SELECT ?s ?p ?o WHERE {{
                            <{s_p}> ?predicate ?s .
                            ?s ?p ?o .
                        }}""",  # request info about children
                        f"""SELECT ?s ?p ?o WHERE {{
                        ?s ?predicate <{s_p}> .
                        ?s ?p ?o .
                        }}""",  # request info about parent
                        f"""SELECT ?s ?p ?o WHERE {{
                        ?parent ?predicate <{s_p}> .
                        ?parent ?another_predicate ?s .
                        ?s ?p ?o .
                        }}""",  # request info about neighbors
                    )
                else:
                    queries = tuple()

                start = datetime.now()
                for query in queries:
                    query_iterator = iter(
                        base.query(
                            query,
                            initNs={
                                "owl": URIRef("http://www.w3.org/2002/07/owl#")
                            },
                        )
                    )

                    result = set()
                    taken = True
                    timeout = False
                    while taken and not timeout:
                        taken = set(take(query_iterator, self._cache_step))
                        result |= taken

                        if datetime.now() - start > self._cache_time:
                            timeout = True
                            break

                    if timeout:
                        break

                    result = set(result)
                    if self._compute_space(older_than=timestamp) >= len(
                        result
                    ):
                        pattern_dict = dict()
                        for s, p, o in result:
                            pattern_dict[(s, None, None)] = pattern_dict.get(
                                (s, None, None), set()
                            ) | {(s, p, o)}

                        for pattern, triples in pattern_dict.items():
                            self._fill(pattern, triples)

        # Pool existing triples minus added and deleted triples.
        triple_pool = query_method(triple_pattern)
        if not ignore_buffers:
            triple_pool = (
                triple
                for triple in triple_pool
                if all(
                    triple not in graph
                    for graph in (
                        self._buffer_uncaught[BufferType.DELETED],
                        self._buffer_caught[BufferType.DELETED],
                        self._buffer_uncaught[BufferType.ADDED],
                        self._buffer_caught[BufferType.ADDED],
                    )
                )
            )
            # Include added triples (previously they were excluded in order not
            # to duplicate triples).
            triple_pool = chain(
                triple_pool,
                self._buffer_uncaught[BufferType.ADDED].triples(
                    triple_pattern
                ),
                self._buffer_caught[BufferType.ADDED].triples(triple_pattern),
            )

        yield from ((triple, iter(())) for triple in triple_pool)

    def __len__(self, context: Graph = None, ignore_buffers=False) -> int:
        """Get the number of triples in the store."""
        # TODO: faster algorithm.
        return sum(
            1 for _ in self.triples((None, None, None), ignore_buffers=True)
        )

    def bind(self, prefix, namespace):
        """Bind a namespace to a prefix."""
        return super().bind(prefix, namespace)

    def namespace(self, prefix):
        """Get the namespace to which a prefix is bound."""
        return super().namespace(prefix)

    def prefix(self, namespace):
        """Get a bound namespace's prefix."""
        return super().prefix(namespace)

    def namespaces(self):
        """Get the bound namespaces."""
        return super().namespaces()

    def query(
        self, query, init_ns, init_bindings, query_graph, **kwargs
    ) -> Result:
        """Perform a SPARQL query on the store."""
        if (
            sum(
                len(x)
                for x in (
                    buffer[y]
                    for buffer in (self._buffer_uncaught, self._buffer_caught)
                    for y in (BufferType.ADDED, BufferType.DELETED)
                )
            )
            > 0
        ):
            # TODO: raise warning that committing can increase query
            #  performance.
            raise NotImplementedError
        elif hasattr(self.interface, "triples"):
            if hasattr(self.interface, "query"):
                # TODO: translate init_ns and init_bindings.
                return self.interface.query(query)
            else:
                raise NotImplementedError
        else:
            return self.interface.base.query(
                query, initNs=init_ns, initBindings=init_bindings, **kwargs
            )

    def update(self, query, init_ns, init_bindings, query_graph, **kwargs):
        """Perform a SPARQL update query on the store."""
        if (
            sum(
                len(x)
                for x in (
                    buffer[y]
                    for buffer in (self._buffer_uncaught, self._buffer_caught)
                    for y in (BufferType.ADDED, BufferType.DELETED)
                )
            )
            > 0
        ):
            # TODO: raise warning that committing can increase query
            #  performance.
            raise NotImplementedError
        elif hasattr(self.interface, "triples"):
            if hasattr(self.interface, "update"):
                return self.interface.update(query)
            else:
                raise NotImplementedError
        else:
            return self.interface.base.update(
                query, initNs=init_ns, initBindings=init_bindings, **kwargs
            )

    def commit(self) -> None:
        """Commit buffered changes."""
        # Gets the old and new graphs and let the interface access a
        # read-only version of them.
        self.interface.old_graph = Graph(
            store=InterfaceDriver.UnbufferedInterfaceDriver(driver=self)
        )
        self.interface.new_graph = ReadOnlyGraphAggregate([Graph(store=self)])

        # Lets the interface access the buffer of caught triples.
        self.interface.buffer = self._buffer_caught

        # Creates the base, old and new sessions for the interface.
        self.interface.session_base = Session(
            base=self.interface.base, ontology=self._ontology
        )
        self.interface.session_old = Session(
            base=self.interface.old_graph, ontology=self._ontology
        )
        self.interface.session_new = Session(
            base=self.interface.new_graph, ontology=self._ontology
        )

        # Computes updated, added and deleted entities if enabled.
        if self.interface.entity_tracking:
            (
                added_entities,
                updated_entities,
                deleted_entities,
            ) = self._compute_entity_modifications()
            self.interface.added = added_entities
            self.interface.updated = updated_entities
            self.interface.deleted = deleted_entities

        # Calls commit on the interface.
        session = self.interface.session_new
        self.interface.session = session
        try:
            session.lock()
            with session:
                self.interface.commit()
        finally:
            session.unlock()
            self.interface.session_new = None
            self.interface.session = None

        # Copies the uncaught triples from the buffers to the base graph.
        for triple in self._buffer_uncaught[BufferType.DELETED]:
            self.interface.base.remove(triple)
        self.interface.base.addN(
            (s, p, o, self.interface.base)
            for s, p, o in self._buffer_uncaught[BufferType.ADDED]
        )
        self.interface.base.commit()

        # Queue file upload and removal for file objects.
        for s in chain(
            self._buffer_uncaught[BufferType.DELETED][
                : RDF.type : simphony_namespace.File
            ],
            self._buffer_caught[BufferType.DELETED][
                : RDF.type : simphony_namespace.File
            ],
        ):
            self.queue(s, None)
        for URI, file in self._queue.items():
            if file is None:
                if hasattr(self.interface, "delete"):
                    self.interface.delete(URI)
                else:
                    logging.warning(
                        f"Ignoring deletion of file {URI}, as the session "
                        f"does not support deleting files."
                    )
            else:
                if hasattr(self.interface, "save"):
                    self.interface.save(URI, file)
                else:
                    logging.warning(
                        f"File {URI}, will NOT be committed to the session, "
                        f"as it does not support the storage of new files."
                    )
                file.close()

        # Reflect changes from buffers in the cache.
        if self.interface.cache:
            # - remove deleted triples
            for triple in chain(
                self._buffer_uncaught[BufferType.DELETED],
                self._buffer_caught[BufferType.DELETED],
            ):
                self._cache.remove(triple)
            # - add the added triples
            timestamp = datetime.now()
            cacheable_triples = dict()
            for triple in chain(
                self._buffer_uncaught[BufferType.ADDED],
                self._buffer_caught[BufferType.ADDED],
            ):
                compatible_cached_patterns = self._compatible_patterns(
                    triple
                ) & set(self._cached_patterns)
                for pattern in compatible_cached_patterns:
                    cacheable_triples[pattern] = cacheable_triples.get(
                        pattern, set()
                    ) | {triple}
            for pattern, triples in cacheable_triples.items():
                try:
                    self._fill(
                        pattern, triples, older_than=timestamp, replace=False
                    )
                except CacheSizeException:
                    self._cache.remove(pattern)
                    try:
                        del self._cached_patterns[pattern]
                    except KeyError:
                        pass

        # Reset buffers and file queue.
        self._buffer_uncaught = {
            buffer_type: Graph(SimpleMemory()) for buffer_type in BufferType
        }
        self._buffer_caught = {
            buffer_type: Graph(SimpleMemory()) for buffer_type in BufferType
        }
        self._queue = dict()

        # Reset graphs, sessions and entities passed to the interface.
        self.interface.old_graph = None
        self.interface.new_graph = None
        self.interface.buffer = None
        self.interface.session_base = None
        self.interface.session_old = None
        self.interface.session_new = None
        self.interface.added = None
        self.interface.updated = None
        self.interface.deleted = None

    def rollback(self) -> None:
        """Discard uncommitted changes."""
        self._buffer_uncaught = {
            buffer_type: Graph(SimpleMemory()) for buffer_type in BufferType
        }
        self._buffer_caught = {
            buffer_type: Graph(SimpleMemory()) for buffer_type in BufferType
        }
        for file in self._queue.values():
            if file is not None:
                file.close()
        self._queue = dict()

    # RDFLib
    # ↑ -- ↑

    def compute(
        self,
        **kwargs: Union[
            str,
            int,
            float,
            bool,
            None,
            Iterable[Union[str, int, float, bool, None]],
        ],
    ):
        """Compute new information (e.g. run a simulation)."""
        if not hasattr(self.interface, "compute"):
            raise AttributeError(
                f"'{self.interface}' object has no attribute 'compute'"
            )

        self.commit()

        self.interface.session_base = Session(
            base=self.interface.base, ontology=self._ontology
        )
        session = self.interface.session_base
        self.interface.session = session
        try:
            session.lock()
            with session:
                self.interface.compute(**kwargs)
                session.commit()
        finally:
            session.unlock()
            self.interface.session_base = None
            self.interface.session = None

    def queue(self, key: URIRef, file: Optional[BinaryIO]) -> None:
        """Queue a file to be committed."""
        if not hasattr(self.interface, "save"):
            logger.warning(
                "This session does not support saving new files. The "
                "contents of the file will NOT be saved during the commit "
                "operation."
            )

        # Clear cached bytestream
        file_name = b64encode(bytes(key, encoding="UTF-8")).decode("UTF-8")
        file_path = Path(self._file_cache.name) / file_name
        if file_path.exists():
            file_path.unlink()

        self._queue[key] = file

    def load(self, key: URIRef) -> BinaryIO:
        """Retrieve a file."""
        if key in self._queue:
            file_name = b64encode(bytes(key, encoding="UTF-8")).decode("UTF-8")

            # Save a temporary copy of the file and put a file handle
            # pointing to the copy on the queue
            if not (Path(self._file_cache.name) / file_name).exists():
                queued = self._queue[key]
                with open(
                    Path(self._file_cache.name) / file_name, "wb"
                ) as file:
                    file.write(queued.read())

                file = open(Path(self._file_cache.name) / file_name, "rb")
                self._queue[key] = file

            # Return a file handle pointing to the copy
            byte_stream = open(Path(self._file_cache.name) / file_name, "rb")
        elif hasattr(self.interface, "load"):
            byte_stream = self.interface.load(key)
        else:
            raise FileNotFoundError(
                "This session does not support file storage. Unable to "
                "retrieve the file contents."
            )

        return byte_stream

    def cache_clear(self):
        """Clear the interface's cache."""
        self._cache.remove((None, None, None))
        self._cached_patterns.clear()

    def _compute_entity_modifications(
        self,
    ) -> Tuple[Set[OntologyEntity], Set[OntologyEntity], Set[OntologyEntity]]:

        # Find out from the triples which entities were added, updated and
        # deleted.
        class Tracker:
            """Checks whether entities already exist on the interface.

            In addition, it keeps track of entities for which existence has
            been already checked.
            """

            existing_subjects: set

            @property
            def visited_subjects(self) -> set:
                return (
                    self.visited_subjects_minus_existing_subjects
                    | self.existing_subjects
                )

            _interface = self.interface

            def __init__(self):
                self.existing_subjects = set()
                self.visited_subjects_minus_existing_subjects = set()

            def __call__(self, subject: Node):
                if subject not in self.visited_subjects:
                    if (
                        next(
                            self._interface.old_graph.triples(
                                (subject, None, None)
                            ),
                            None,
                        )
                        is not None
                    ):
                        self.existing_subjects.add(subject)
                    else:
                        self.visited_subjects_minus_existing_subjects.add(
                            subject
                        )

        tracker = Tracker()

        # Get added subjects.
        for s in chain(
            self._buffer_uncaught[BufferType.ADDED].subjects(),
            self._buffer_caught[BufferType.ADDED].subjects(),
        ):
            tracker(s)
        added_subjects = tracker.visited_subjects_minus_existing_subjects
        # Get deleted subjects.
        deleted_subjects = dict()
        for s in chain(
            self._buffer_uncaught[BufferType.DELETED].subjects(),
            self._buffer_caught[BufferType.DELETED].subjects(),
        ):
            tracker(s)
            if s not in added_subjects and s in tracker.existing_subjects:
                deleted_subjects[s] = deleted_subjects.get(s, 0) + 1
        deleted_subjects = {
            s: True
            if count
            >= sum(
                1 for _ in self.interface.old_graph.triples((s, None, None))
            )
            else False
            for s, count in deleted_subjects.items()
        }
        deleted_subjects = {
            s for s, deleted in deleted_subjects.items() if deleted
        }
        # Get updated subjects.
        updated_subjects = tracker.existing_subjects.difference(
            deleted_subjects
        )
        added_entities = {
            self.interface.session_new.from_identifier(s)
            for s in added_subjects
        }
        updated_entities = {
            self.interface.session_new.from_identifier(s)
            for s in updated_subjects
        }
        deleted_entities = {
            self.interface.session_old.from_identifier(s)
            for s in deleted_subjects
        }
        return added_entities, updated_entities, deleted_entities

    class UnbufferedInterfaceDriver(Store):
        """Provides a read-only view on the interface without the buffers.

        This store implementation provides a read-only view on the connected
        interface without the `InterfaceDriver`'s buffers.
        """

        _driver: InterfaceDriver

        # RDFLib
        # ↓ -- ↓

        context_aware: bool = False
        formula_aware: bool = False
        transaction_aware: bool = True
        graph_aware: bool = False

        def __init__(self, *args, driver: InterfaceDriver, **kwargs):
            """Initialize the store."""
            self._driver = driver
            super().__init__(*args, **kwargs)

        def open(self, configuration: str, create: bool = False) -> None:
            """The interface should be already open. Prevents reopening it."""
            raise ModificationException

        def close(self, commit_pending_transaction=False) -> None:
            """The interface should be open. Prevents closing it."""
            raise ModificationException

        def add(self, triple: Triple, context: Graph, quoted=False) -> None:
            """The view is read-only, prevents modifications."""
            raise ModificationException

        def remove(
            self, triple_pattern: Pattern, context: Optional[Graph] = None
        ) -> None:
            """The view is read-only, prevents modifications."""
            raise ModificationException

        def triples(
            self, triple_pattern: Pattern, context=None
        ) -> Iterator[Tuple[Triple, Graph]]:
            """Yields triples from the interface ignoring the buffers."""
            yield from self._driver.triples(
                triple_pattern, context, ignore_buffers=True
            )

        def __len__(self, context: Graph = None) -> int:
            """Gets the amount of triples on the interface, without buffers."""
            return self._driver.__len__(context, ignore_buffers=True)

        def bind(self, prefix, namespace):
            """Prevents the modification of namespaces."""
            raise ModificationException

        def namespace(self, prefix):
            """Get the namespace to which a prefix is bound."""
            return self._driver.namespace(prefix)

        def prefix(self, namespace):
            """Get a bound namespace's prefix."""
            return self._driver.prefix(namespace)

        def namespaces(self):
            """Get the bound namespaces."""
            return self._driver.namespaces()

        def query(self, *args, **kwargs) -> Result:
            """Perform a SPARQL query on the store."""
            return self._driver.query(*args, **kwargs)

        def update(self, *args, **kwargs):
            """Prevents the modification."""
            raise ModificationException

    def _fill(
        self,
        triple_pattern: Pattern,
        triples: Collection[Triple],
        older_than: Optional[datetime] = None,
        replace: bool = True,
    ) -> None:
        """Fill the cache with new triples.

        Args:
            triples: An collection of triples to fill the cache with.
            triple_pattern: The triple pattern that is being updated.
            older_than: Cache entries older than this date can be cleared
                to make room for the new triples. When not specified,
                `datetime.now()` is invoked.

        Raises:
            CacheSizeException: No space left on the cache to fill it with
            the provided triples.
        """
        if older_than is None:
            older_than = datetime.now()

        """Check if there is room for the new triples.

        Also compute the patterns that should be removed to make room for
        the new one.
        """
        if len(triples) > self._cache_size:
            """The triples to store are bigger than the cache."""
            raise CacheSizeException()
        elif len(triples) > (self._cache_size - len(self._cache)):
            """Not enough free space in the cache.

            Find out which cached patterns can be removed to make room for
            the new triples.
            """
            to_remove = set()
            removable_patterns = dict()
            # sub-patterns of the pattern to cache are always removable
            removable_patterns.update(
                {
                    pattern: None
                    for pattern in set(self._cached_patterns)
                    - {triple_pattern}
                    if self._is_sub_pattern(pattern, triple_pattern)
                }
            )
            if replace:
                removable_patterns.update({triple_pattern: None})
            to_remove |= set(removable_patterns)
            # concatenate the rest of the patterns:
            # TODO: do not sort, keep ordering when inserting
            removable_patterns.update(
                {
                    pattern: date
                    for pattern, date in sorted(
                        self._cached_patterns.items(), key=lambda x: x[1]
                    )
                    if date < older_than
                }
            )
            triple_count = 0
            for pattern in removable_patterns:
                triple_count += sum(1 for _ in self._cache.triples(pattern))
                to_remove.add(pattern)
                if triple_count >= len(triples):
                    break
            else:
                raise CacheSizeException()

            """Remove the patterns to make room for the new one."""
            for pattern in to_remove:
                self._cache.remove(pattern)
                try:
                    del self._cached_patterns[pattern]
                except KeyError:
                    pass

        """Fill the cache with the new pattern."""
        self._cache.remove(triple_pattern)
        self._cache.addN((s, p, o, self._cache) for s, p, o in triples)
        self._cached_patterns[triple_pattern] = datetime.now()

    def _compute_space(
        self,
        older_than: Optional[datetime] = None,
    ) -> int:
        """Compute the free space on the cache.

        Args:
            older_than: Consider patterns cached more recently than this
                date as pinned.

        Returns:
            Free space on the cache (in triples).
        """
        if older_than is None:
            older_than = datetime.now()

        patterns = {
            pattern
            for pattern, date in self._cached_patterns.items()
            if date < older_than
        }
        triple_count = 0
        for pattern in patterns:
            triple_count += sum(1 for _ in self._cache.triples(pattern))

        return min(
            self._cache_size,
            self._cache_size - len(self._cache) + triple_count,
        )

    @staticmethod
    def _is_sub_pattern(sub_pattern: Pattern, pattern: Pattern) -> bool:
        """Determine whether a triple pattern is a sub-pattern of another.

        Args:
            sub_pattern: The pattern that is assumed to be a sub-pattern.
            pattern: The pattern that is assumed to be the super-pattern.

        Returns:
            Whether the above assumptions are true or not.
        """
        return all(
            sub_pattern[i] == pattern[i] or pattern[i] is None
            for i in range(0, 3)
        )

    @staticmethod
    def _compatible_patterns(triple: Triple) -> Set[Pattern]:
        """List the patterns that are compatible with a given triple."""
        s, p, o = triple
        return {
            (s, p, o),
            (s, p, None),
            (None, p, o),
            (s, None, o),
            (None, None, o),
            (s, None, None),
            (None, p, None),
            (None, None, None),
        }

    def _cached(self, item: Pattern) -> bool:
        """Determine whether a given pattern is already cached."""
        return bool(
            self._compatible_patterns(item) & set(self._cached_patterns)
        )


class Interface(ABC):
    """To be implemented by interface/wrapper developers.

    This is the most generic type of interface.
    """

    # Definition of:
    # Interface
    # ↓ ----- ↓

    base: Optional[Graph] = None
    """A base graph can be provided by the wrapper developer."""

    entity_tracking: bool = True
    """Whether to enable entity tracking.

    With entity tracking enabled, a list of added, updated and deleted
    entities will be available during commit. If such lists are not
    going to be used, then this property can be set to `False` to
    increase performance.
    """

    cache: bool = False
    """Whether to enable caching.

    When caching is enabled, the number of calls to the methods of the
    interface are reduced. This is useful if the latency of such calls is
    high. This is the case, for example,
    """

    def __init__(
        self,
        **kwargs: Union[
            str,
            int,
            float,
            bool,
            None,
            Iterable[Union[str, int, float, bool, None]],
        ],
    ):
        """Initialize the interface.

        The `__init__` method accepts JSON-serializable keyword arguments
        in order to let the user configure parameters of the interface that
        are not configurable via the ontology. For example, the type of
        solver used by a simulation engine.

        Args:
            kwargs: JSON-serializable keyword arguments that contain no
                nested JSON objects (check the type hint for this argument).
        """
        super().__init__()

    @abstractmethod
    def open(
        self,
        configuration: str,
        create: bool = False,
    ) -> None:
        """Open the data source that the interface interacts with.

        You can expect calls to this method even when the data source is
        already open, therefore, an implementation like the following is
        recommended.

        def open(self, configuration: str, create: bool = False):
            if your_data_source_is_already_open:
                return
                # To improve the user experience you can check if the
                # configuration string leads to a resource different from
                # the current one and raise an error informing the user.

            # Connect to your data source...
            # your_data_source_is_already_open is for now True.

        If you are using a custom base graph (for example based on an RDFLib
        store), please set `self.base = your_graph` within this method.
        Otherwise, an empty base graph will be provided.

        Args:
            configuration: Determines the location of the data source to be
                opened.
            create: Whether the data source can be created at the target
                location if it does not exist. When false, if the data
                source does not exist, you should raise an exception. When
                true, create an empty data source.
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """Close the data source that the interface interacts with.

        This method should NOT commit uncommitted changes.

        This method should close the connection that was obtained in `open`,
        and free any locked up resources.

        You can expect calls to this method even when the data source is
        already closed. Therefore, an implementation like the following is
        recommended.

        def close(self):
            if your_data_source_is_already_closed:
                return

            # Close the connection to your data source.
            # your_data_source_is_already_closed is for now True
        """
        pass

    @abstractmethod
    def populate(self) -> None:
        """Populate the base graph so that it represents the data source.

        This command is run after the data source is opened. Here you are
        expected to populate the base graph so that its information mimics
        the information on the data source, unless you are generating
        triples on the fly. The default session is a session based on the
        base graph.

        The base graph is available on `self.base`, and a session based on
        the base graph is available on `self.session` and `self.session_base`.
        """
        pass

    @abstractmethod
    def commit(self) -> None:
        """This method commits the changes made by the user.

        Here, you are expected to have access to the following:
        - `self.base`: The base graph (rw). You are not expected to modify it.
        - `self.old_graph`: The old graph (ro).
        - `self.new_Graph`: The new graph (ro).
        - `self.buffer`: The buffer of caught triples (rw) that you now have to
            reflect on the data structures of your software.
        - `self.session_base`: A session based on the base graph (rw). You
            are not expected to modify it.
        - `self.session_old`: A session based on the old graph (ro).
        - `self.session_new`: A session based on the new graph (ro).
        - `self.session`: same as `self.session_new`.
        - `self.added`: A list of added entities (rw). You are not expected
            to modify the entities.
        - `self.updated`: A list of updated entities (rw). You are not
            expected to modify the entities.
        - `self.deleted`: A list of deleted entities (rw). You are not
            expected to modify the entities.

        Raises:
            AssertionError: When the data provided by the user would produce
                an inconsistent or unpredictable state of the data structures.
        """
        # Before updating the data structures, check that the changes provided
        # by the user do not leave them in a consistent state. This necessary
        # because SimPhoNy cannot revert the changes you make to your
        # data structures. Raise an AssertionError if the check fails.

        # Examine the differences between the graphs below and make a plan to
        # modify your data structures.

        # Change your data structures below. As you change the data
        # structures, `old_graph` and `new_graph` will change, so do not
        # rely on them. Relying on them would be analogous to changing a
        # dictionary while looping over it.
        pass

    def compute(
        self,
        **kwargs: Union[
            str,
            int,
            float,
            bool,
            None,
            Iterable[Union[str, int, float, bool, None]],
        ],
    ) -> None:
        """Compute new information (e.g. run a simulation).

        Just compute the new information on the backend and reflect the
        changes on the base graph. The default session is the base session.

        The base graph is available on `self.base`, and a session based on
        the base graph is available on `self.session` and `self.session_base`.
        """
        pass

    del compute  # By default not defined.

    # Triplestore methods.

    def add(self, triple: Triple) -> bool:
        """Inspect and control the addition of triples to the base graph.

        Args:
            triple: The triple being added.

        Returns:
            - True: The triple should be added to the base graph.
            - False: The triple should be caught, and therefore not added to
                the base graph. This triple will be latter available during
                commit on the buffer so that the changes that it introduces
                can be translated to the data structure.
        """
        pass

    del add  # By default not defined.

    def remove(self, pattern: Pattern) -> Iterator[Triple]:
        """Inspect and control the removal of triples from the base graph.

        Args:
            pattern: The pattern being removed.

        Returns:
            An iterator with the triples that should be removed from the
            base graph. Any triples not included will not be removed,
            and will be available on the buffer during commit.
        """
        pass

    del remove  # By default not defined.

    def triples(self, pattern: Pattern) -> Iterator[Triple]:
        """Intercept a triple pattern query.

        Can be used to produce triples that do not exist on the base graph
        on the fly.

        Args:
            pattern: The pattern being queried.

        Returns:
            An iterator yielding triples
        """
        pass

    del triples

    # File storage methods.

    def save(self, key: str, file: BinaryIO) -> None:
        """Save a file."""
        pass

    del save  # By default not defined.

    def load(self, key: str) -> BinaryIO:
        """Retrieve a file."""
        pass

    del load  # By default not defined.

    def delete(self, key: str) -> None:
        """Delete a file."""
        pass

    del delete  # By default not defined.

    def hash(self, key: str) -> str:
        """Hash a file."""
        pass

    del hash  # By default not defined.

    def rename(self, key: str, new_key: str) -> None:
        """Rename a file."""
        pass

    del rename  # By default not defined.

    # The properties below are set by the driver and accessible on the
    # interface. They are not meant to be set by the developers.
    old_graph: Optional[Graph] = None
    new_graph: Optional[Graph] = None
    buffer: Optional[Graph] = None

    session_base: Optional[Session] = None
    session_old: Optional[Session] = None
    session_new: Optional[Session] = None
    session: Optional[Session] = None
    added: Optional[Set[OntologyEntity]] = None
    updated: Optional[Set[OntologyEntity]] = None
    deleted: Optional[Set[OntologyEntity]] = None

    # Definition of:
    # Interface
    # ↑ ----- ↑



"""Utility resources for the ontology module."""
from __future__ import annotations

import importlib
import os
import pkgutil
from abc import ABC, abstractmethod
from functools import lru_cache, reduce
from typing import Any, Dict, Iterable, Iterator, MutableSet, Set, Tuple

from simphony_osp.ontology.entity import OntologyEntity


class DataStructureSet(ABC, MutableSet):
    """Set-like object that acts as an interface to another data structure.

    This class looks like and acts like the standard `set`, but the data it
    holds is stored on a different data structure. When an instance is read
    or when it is modified in-place, specific methods defined to interact
    with the data structure are called to reflect the changes.

    This class does not hold any object-related information itself, thus
    it is safe to spawn multiple instances linked to the same data
    structure (when single-threading).
    """

    @abstractmethod
    def __iter__(self) -> Iterator[Any]:
        """Implement iter(self).

        Returns the associated values from the connected data structure.

        The values provided MUST NOT be repeated, as instances of this class
        are expected to behave like sets. Remember that the definition of
        membership for sets in Python is `any(x is e or x == e for e in y)`,
        see https://docs.python.org/3/reference/expressions.html
        #membership-test-details. Please make sure this function returns
        unique items according to the definition above.

        Returns:
            An iterator providing the values from the connected data
            structure.
        """
        pass

    @abstractmethod
    def __contains__(self, item: Any) -> bool:
        """Return y in x.

        Check if the item is in the connected data structure.

        You can also choose to just call `return super().__contains__(item)` to
        use this naive implementation instead of implementing something
        yourself.
        """
        for x in self:
            if x is item or x == item:
                return True
        return False

    @abstractmethod
    def update(self, other: Iterable) -> None:
        """Update a set with the union of itself and others.

        Updates the connected data structure with the result of performing a
        set union operation with another iterable.
        """
        pass

    @abstractmethod
    def intersection_update(self, other: Iterable) -> None:
        """Update a set with the intersection of itself and another.

        Updates the connected data structure performing a set intersection
        operation with another iterable.
        """
        pass

    @abstractmethod
    def difference_update(self, other: Iterable) -> None:
        """Remove all elements of another set from this set.

        Updates the connected data structure performing a set difference
        operation with another iterable.
        """
        pass

    @abstractmethod
    def symmetric_difference_update(self, other: Iterable) -> None:
        """Update a set with the symmetric difference of itself and another.

        Updates the connected data structure performing an XOR set operation
        with another iterable.
        """
        pass

    def __repr__(self) -> str:
        """Return repr(self)."""
        return set(self).__repr__()

    def __str__(self) -> str:
        """Return str(self)."""
        return set(self).__str__()

    def __format__(self, format_spec: str) -> str:
        """Default object formatter."""
        return set(self).__format__(format_spec)

    def __len__(self) -> int:
        """Return len(self)."""
        return sum(1 for _ in self)

    def __le__(self, other: set) -> bool:
        """Return self<=other."""
        return set(self).__le__(other)

    def __lt__(self, other: set) -> bool:
        """Return self<other."""
        return set(self).__lt__(other)

    def __eq__(self, other: set) -> bool:
        """Return self==other."""
        return set(self).__eq__(other)

    def __ne__(self, other: set) -> bool:
        """Return self!=other."""
        return set(self).__ne__(other)

    def __gt__(self, other: set) -> bool:
        """Return self>other."""
        return set(self).__gt__(other)

    def __ge__(self, other: set) -> bool:
        """Return self>=other."""
        return set(self).__ge__(other)

    def __and__(self, other: set) -> set:
        """Return self&other."""
        return set(self).__and__(other)

    def __radd__(self, other: set) -> set:
        """Return other&self."""
        return other & set(self)

    def __ror__(self, other: set) -> set:
        """Return other|self."""
        return other | set(self)

    def __rsub__(self, other: set) -> set:
        """Return value-self."""
        return other - set(self)

    def __rxor__(self, other: set) -> set:
        """Return value^self."""
        return other ^ set(self)

    def __or__(self, other: set) -> set:
        """Return self|other."""
        return set(self).__or__(other)

    def __sub__(self, other: set) -> set:
        """Return self-other."""
        return set(self).__sub__(other)

    def __xor__(self, other: set) -> set:
        """Return self^other."""
        return set(self).__xor__(other)

    def __iadd__(self, other: Any) -> DataStructureSet:
        """Return self+=other (equivalent to self|=other)."""
        if isinstance(other, (Set, MutableSet)):
            # Apparently instances of MutableSet are not instances of Set.
            self.update(other)
        else:
            self.update({other})
        return self

    def __isub__(self, other: Any) -> DataStructureSet:
        """Return self-=other.

        Based on `difference_update`.
        """
        if isinstance(other, (Set, MutableSet)):
            # Apparently instances of MutableSet are not instances of Set.
            self.difference_update(other)
        else:
            self.difference_update({other})
        return self

    def __ior__(self, other: set) -> DataStructureSet:
        """Return self|=other.

        Should perform the union on the underlying data structure.
        """
        self.update(other)
        return self

    def __iand__(self, other: set) -> DataStructureSet:
        """Return self&=other.

        Should perform the intersection on the underlying data structure.
        """
        self.intersection_update(other)
        return self

    def __ixor__(self, other: set) -> DataStructureSet:
        """Return self^=other.

        Should perform the XOR operation on the underlying data structure.
        """
        self.symmetric_difference_update(other)
        return self

    def isdisjoint(self, other: set) -> bool:
        """Return True if two sets have a null intersection."""
        return set(self).isdisjoint(other)

    def clear(self) -> None:
        """Remove all elements from this set."""
        self.intersection_update(set())

    def pop(self) -> Any:
        """Remove and return an arbitrary set element.

        Raises KeyError if the set is empty.
        """
        try:
            item = next(iter(self))
        except StopIteration:
            return set().pop()  # Underlying data structure is empty.
            # The `KeyError` is raised like this intentionally, to avoid
            #  hardcoding the error text for a set.
        self.difference_update({item})
        return item

    def copy(self) -> set:
        """Return a shallow copy of a set."""
        return set(self)

    def difference(self, other: Iterable) -> set:
        """Return the difference of two or more sets as a new set.

        (i.e. all elements that are in this set but not the others.)
        """
        return set(self).difference(other)

    def discard(self, other: Any) -> None:
        """Remove an element from a set if it is a member.

        If the element is not a member, do nothing.
        """
        self.difference_update({other})

    def intersection(self, other: set) -> set:
        """Return the intersection of two sets as a new set.

        (i.e. all elements that are in both sets.)
        """
        return set(self).intersection(other)

    def issubset(self, other: set) -> bool:
        """Report whether another set contains this set."""
        return self <= other

    def issuperset(self, other: set) -> bool:
        """Report whether this set contains another set."""
        return self >= other

    def add(self, other: Any) -> None:
        """Add an element to a set.

        This has no effect if the element is already present.
        """
        self.update({other})

    def remove(self, other: Any) -> None:
        """Remove an element from a set; it must be a member.

        If the element is not a member, raise a KeyError.
        """
        if other not in self:
            raise KeyError(f"{other}")
        self.difference_update({other})

    def symmetric_difference(self, other: set) -> set:
        """Return the symmetric difference of two sets as a new set."""
        return set(self).symmetric_difference(other)

    def union(self, other: set) -> set:
        """Return the union of sets as a new set."""
        return set(self).union(other)


"""Define a `compatible_classes` function that lists the Python classes that
can be spawned for a particular RDF.type and type of node identifier.

This function used by `simphony_osp.session.session.Session.from_identifier` to
determine which Python class should be spawned for a given IRI or BNode.

Read the docstrings of the functions defined in this section for more details.
"""
# ↓ --------------- ↓


@lru_cache(maxsize=None)
def _compute_mappings() -> Tuple[Dict[Any, Any], Dict[Any, Any]]:
    """Maps RDF types and node identifier types to Python classes.

    The classes defined in SimPhoNy that are meant to represent ontology
    entities (all subclasses of `OntologyEntity`), have two attributes,
    `rdf_type` and `rdf_identifier` that determine the combination of RDF
    types (e.g. owl:Class) and node identifier types (e.g. URIRef, BNode) that
    the class is meant to represent.

    This function imports the `simphony_osp.ontology` module and all of
    its submodules recursively, as all subclasses of `OntologyEntity` are
    expected to be stored there.

    After that, it generates two mappings, `mapping_rdf_to_python_class` and
    `mapping_identifier_to_python_class`, that map each rdf type and node
    identifier type to compatible Python classes.

    Using such mappings, other functions can find out the compatible classes
    for a specific pair of RDF type and node identifier type.

    This function is cached, as it is called by `compatible_classes`
    repeatedly, but the computation is only needed once. In fact, this code
    could be defined outside a function, but it has been incorporated
    into a function because of the need to evaluate it lazily (to avoid
    circular imports).

    Returns:
        A tuple `mapping_rdf_to_python_class,
        mapping_identifier_to_python_class` containing the aforementioned
        mappings.
    """
    # First, import the ontology module and all of its submodules recursively.
    self = __import__(__name__)
    package_paths = [
        os.path.abspath(os.path.join(path, "ontology"))
        for path in self.__path__
    ]
    package_prefix = f"{self.__name__}.ontology."

    def import_modules_recursively(paths, prefix):
        for module_info in pkgutil.iter_modules(paths, prefix):
            module = importlib.import_module(module_info.name)
            if module_info.ispkg:
                import_modules_recursively(
                    module.__path__, f"{module_info.name}."
                )

    import_modules_recursively(package_paths, package_prefix)

    # Then compute the mappings, remember that the python class to instantiate
    # for a given ontology entity depends on two things:
    # - The RDF.type(s) of the identifier.
    # - The RDF node type of the identifier (URIRef, Node or Literal)

    mapping_rdf_to_python_class = dict()
    mapping_identifier_to_python_class = dict()

    def recursive_iterator(class_):
        for sub_class in class_.__subclasses__():
            yield from recursive_iterator(sub_class)
            yield sub_class

    for subclass in recursive_iterator(OntologyEntity):
        rdf_types = (
            subclass.rdf_type
            if isinstance(subclass.rdf_type, set)
            else {subclass.rdf_type}
        )
        for rdf_type in rdf_types:
            mapping_rdf_to_python_class[
                rdf_type
            ] = mapping_rdf_to_python_class.get(rdf_type, set()) | {subclass}
        rdf_identifiers = (
            subclass.rdf_identifier
            if isinstance(subclass.rdf_identifier, set)
            else {subclass.rdf_identifier}
        )
        for rdf_identifier in rdf_identifiers:
            mapping_identifier_to_python_class[
                rdf_identifier
            ] = mapping_identifier_to_python_class.get(
                rdf_identifier, set()
            ) | {
                subclass
            }
    return mapping_rdf_to_python_class, mapping_identifier_to_python_class


@lru_cache(maxsize=4096)
def compatible_classes(type_, identifier):
    """Given a pair of an RDF type and an identifier get a Python class.

    Given a pair of an RDF type and an identifier, the compatible Python
    classes are computed. In fact, for the latter only the type of
    identifier matters.
    """
    (
        mapping_rdf_to_python_class,
        mapping_identifier_to_python_class,
    ) = _compute_mappings()
    # Remember that the call above is cached (see `_compute_mappings`).

    from_type = mapping_rdf_to_python_class.get(type_, set())
    from_identifier = reduce(
        lambda x, y: x | y,
        (
            value
            for key, value in mapping_identifier_to_python_class.items()
            if isinstance(identifier, key)
        ),
        set(),
    )
    return from_type & from_identifier


# ↑ --------------- ↑



"""SimPhoNy OSP module."""

from importlib import import_module as _import_module

_import_module("simphony_osp.utils.pico")  # Load installed ontologies

__all__ = [
    "ontology",
    "session",
    "tools",
    "development",
    "namespaces",
    "wrappers",
]



"""Pico is a commandline tool used to install ontologies."""

import argparse
import logging
from pathlib import Path
from typing import Tuple, Union

from simphony_osp.ontology.namespace import OntologyNamespace
from simphony_osp.utils.pico import logger, pico

__all__ = ["packages", "namespaces", "install", "uninstall"]


def packages() -> Tuple[str]:
    """Returns the identifiers of all installed packages."""
    return pico.packages


def namespaces() -> Tuple[OntologyNamespace]:
    """Returns namespace objects for all the installed namespaces."""
    return pico.namespaces


def install(*files: Union[Path, str], overwrite: bool = False) -> None:
    """Install ontology packages.

    Args:
        files: Paths of `yml` files describing the ontologies to install.
            Alternatively, identifiers of ontology packages that are
            bundled with SimPhoNy.
        overwrite: Whether to overwrite already installed ontology
            packages.
    """
    return pico.install(*files, overwrite=overwrite)


def uninstall(*identifiers: str) -> None:
    """Uninstall ontology packages.

    Args:
        identifiers: Identifiers of the ontology packages to uninstall.
    """
    return pico.uninstall(*identifiers)


def terminal() -> None:
    """Install ontologies from the terminal."""
    # Entry point
    parser = argparse.ArgumentParser(
        description="This tool manages the ontologies used by SimPhoNy."
    )
    parser.add_argument(
        "--log-level",
        default="INFO",
        type=str.upper,
        help="set the logging level",
    )

    # --- Available commands ---
    subparsers = parser.add_subparsers(title="command", dest="command")

    # list
    subparsers.add_parser("list", help="Lists all the installed ontologies")

    # install
    install_parser = subparsers.add_parser(
        "install", help="Install ontologies"
    )
    install_parser.add_argument(
        "files", nargs="+", type=str, help="List of yaml files to install"
    )
    install_parser.add_argument(
        "--overwrite",
        action="store_true",
        help="Overwrite the existing ontologies, if they are already "
        "installed",
    )

    # uninstall parser
    uninstall_parser = subparsers.add_parser(
        "uninstall", help="Uninstall ontologies"
    )
    uninstall_parser.add_argument(
        "packages",
        nargs="+",
        type=str,
        help="List of ontology packages to uninstall",
    )

    args = parser.parse_args()
    logging.getLogger("simphony_osp").setLevel(
        getattr(logging, args.log_level)
    )

    try:
        installed_identifiers = list(pico.packages)
        if args.command == "install" and args.overwrite:
            pico.install(*args.files, overwrite=True)
        elif args.command == "install":
            pico.install(*args.files, overwrite=False)
        elif args.command == "uninstall":
            if args.packages == ["all"]:
                args.packages = installed_identifiers
            pico.uninstall(*args.packages)
        elif args.command == "list":
            print("Packages:")
            print("\n".join(map(lambda x: "\t- " + x, installed_identifiers)))

            installed_namespaces = tuple(
                namespace.name for namespace in pico.ontology.namespaces
            )
            print("Namespaces:")
            print("\n".join(map(lambda x: "\t- " + x, installed_namespaces)))
    except Exception:
        logger.error("An Exception occurred during installation.", exc_info=1)
        if args.log_level != "DEBUG":
            logger.error(
                "Consider running 'pico --log-level debug %s ...'"
                % args.command
            )


if __name__ == "__main__":
    terminal()



"""Visualize an ontology, an individual or a session using graphviz."""

import argparse
import logging
import os
from pathlib import Path
from typing import FrozenSet, Iterable, Optional, Union
from uuid import UUID

from graphviz import Digraph

from simphony_osp.namespaces import owl
from simphony_osp.ontology.attribute import OntologyAttribute
from simphony_osp.ontology.entity import OntologyEntity
from simphony_osp.ontology.individual import OntologyIndividual
from simphony_osp.ontology.namespace import OntologyNamespace
from simphony_osp.ontology.oclass import OntologyClass
from simphony_osp.ontology.parser import OntologyParser
from simphony_osp.ontology.relationship import OntologyRelationship
from simphony_osp.session.session import Session
from simphony_osp.tools.search import find
from simphony_osp.utils.datatypes import UID

logger = logging.getLogger(__name__)


class Semantic2Dot:
    """Utility for creating a dot and png representation of semantic data."""

    def __init__(
        self,
        *elements: Union[
            OntologyIndividual,
            OntologyNamespace,
            Session,
        ],
        rel: Optional[
            Union[OntologyRelationship, Iterable[OntologyRelationship]]
        ] = None,
    ):
        """Initializes the class."""
        if rel is not None:
            rel = {rel} if isinstance(rel, OntologyRelationship) else set(rel)

        # Classify items to be drawn
        classes = {OntologyIndividual, OntologyNamespace, Session}
        class_identified = (
            (class_, element)
            for element in elements
            for class_ in classes
            if isinstance(element, class_)
        )
        classification = {
            OntologyIndividual: set(),
            OntologyNamespace: set(),
            Session: set(),
        }
        for class_, element in class_identified:
            classification[class_].add(element)
        unclassified = set(elements) - set(
            item for items in classification.values() for item in items
        )
        if unclassified:
            raise TypeError(
                f"The object{'s' if len(unclassified) > 2 else ''} "
                f"{','.join(str(x) for x in unclassified)} "
                f"cannot be drawn."
            )

        # Save nodes requested to be drawn
        supported_classes = (
            OntologyIndividual,
            OntologyClass,
            OntologyAttribute,
            OntologyRelationship,
        )
        self._requested_individuals = frozenset(
            classification[OntologyIndividual]
        )
        self._requested = frozenset(
            {
                part
                for individual in classification[OntologyIndividual]
                for part in (
                    find(individual, rel=rel, find_all=True)
                    if rel is not None
                    else (individual,)
                )
            }
            | {
                entity
                for namespace in classification[OntologyNamespace]
                for entity in namespace
                if isinstance(entity, supported_classes)
            }
            | {
                entity
                for session in classification[Session]
                for entity in session
            }
        )

        self._graph = Digraph(format="png", name="SimPhoNy semantic2dot")
        self._draw_all()

    def render(self, filename: str = None, **kwargs) -> None:
        """Save the graph to a dot and png file."""
        if filename is None:
            raise ValueError("Please specify a file name to save your graph.")
        logger.info("Writing file %s" % filename)
        self._graph.render(filename=filename, **kwargs)

    def _repr_mimebundle_(
        self,
        include: Optional[Iterable[str]],
        exclude: Optional[Iterable[str]],
    ):
        """Render the graph as an image on IPython (e.g. Jupyter notebooks)."""
        return self._graph._repr_mimebundle_(include, exclude)

    _label = (
        "<<TABLE BORDER='0' CELLBORDER='0'>"
        "<TR><TD>{}</TD></TR>"
        "{}"
        "</TABLE>>"
    )

    _attribute = "<TR ALIGN='left'><TD>{}: {}</TD></TR>"

    _requested: FrozenSet[OntologyEntity]
    _requested_individuals: FrozenSet[OntologyIndividual]

    def _draw_all(self) -> None:
        """Draws all the requested items.

        Goes over all the requested items, draws them, also draws any
        complementary items (e.g. superclasses) and finally draws the edges.

        REMARK: This method should be run ONLY ONCE.
        """
        # keep track of ontology individual relationships
        edges = {
            "individual_relationships": set(),
            "complementary_superclasses": set(),
            "inverse_relationships": set(),
        }
        nodes = self._requested
        complementary_nodes = set()

        # draw nodes, compute edges and complementary node
        for node in nodes:
            if isinstance(node, OntologyIndividual):
                edges["individual_relationships"] |= {
                    (node, relationship, target)
                    for target, relationship in node.relationships_iter(
                        return_rel=True
                    )
                    if target in nodes
                }
            elif isinstance(node, OntologyRelationship):
                superclasses = set(node.direct_superclasses)
                if not superclasses and owl.topObjectProperty != node:
                    superclasses |= {owl.topObjectProperty}
                complementary_nodes |= superclasses
                edges["complementary_superclasses"] |= {
                    (node, superclass) for superclass in superclasses
                }
                inverse = node.inverse
                if inverse in nodes:
                    edges["inverse_relationships"].add(
                        frozenset({node, inverse})
                    )
            elif isinstance(node, OntologyAttribute):
                superclasses = set(node.direct_superclasses)
                if not superclasses and owl.topDataProperty != node:
                    superclasses |= {owl.topDataProperty}
                complementary_nodes |= superclasses
                edges["complementary_superclasses"] |= {
                    (node, superclass) for superclass in superclasses
                }
            elif isinstance(node, OntologyClass):
                superclasses = set(node.direct_superclasses)
                if not superclasses and owl.Thing != node:
                    superclasses |= {owl.Thing}
                complementary_nodes |= superclasses
                edges["complementary_superclasses"] |= {
                    (node, superclass) for superclass in superclasses
                }

            self._draw_node(node, complementary=False)

        complementary_nodes = complementary_nodes - nodes

        # draw complementary nodes
        for node in complementary_nodes:
            self._draw_node(node, complementary=True)

        # draw edges
        # - direct superclasses
        for node, superclass in edges["complementary_superclasses"]:
            self._draw_edge(node, superclass, label="is_a")
        # - relationships between individuals
        for start, rel, end in edges["individual_relationships"]:
            self._draw_edge(
                start,
                end,
                label=(
                    self._get_element_label(rel) + f" ({rel.namespace.name})"
                    if rel.namespace
                    else ""
                ),
            )
        # - inverse relationships
        for node1, node2 in edges["inverse_relationships"]:
            self._draw_edge(
                node1, node2, label="inverse", dir="none", style="dashed"
            )

    @staticmethod
    def _get_element_label(
        element: Union[
            OntologyEntity,
            Session,
            OntologyNamespace,
        ]
    ) -> str:
        """Compute a label for an ontology entity, session or namespace."""
        if isinstance(element, OntologyEntity):
            name = Semantic2Dot._get_ontology_entity_label(element)
        elif isinstance(element, Session):
            name = f"{hex(id(Session))}"
        elif isinstance(element, OntologyNamespace):
            name = f"{element.name}"
        else:
            raise TypeError(f"Unsupported element type {type(element)}.")
        return str(name)

    @staticmethod
    def _get_ontology_entity_label(element: OntologyEntity) -> str:
        """Compute a label for an ontology entity."""
        # Try label
        name = element.label

        if name is None:
            # Try suffix
            name = (
                element.iri[len(element.namespace.iri) :]
                if element.namespace is not None
                else None
            )

        # Try UUID or identifier
        if name is None:
            if isinstance(element.uid.data, UUID):
                name = Semantic2Dot._shorten_uid(element.uid)
            else:
                name = element.identifier

        return name

    @staticmethod
    def _shorten_uid(uid: UID) -> str:
        """Shortens the string of a given uid.

        Args:
            uid: UID to shorten
        Returns:
            str: 8 first and 3 last characters separated by '...'.
        """
        uid = str(uid)
        return uid[:8] + "..." + uid[-3:]

    def _draw_node(
        self, entity: OntologyEntity, complementary: bool = False
    ) -> None:
        """Choose the method to draw an ontology entity and call it.

        Selects the correct method to draw an ontology entity according to
        its type and calls it.

        Args:
            entity: The ontology entity to draw.

        Raises:
            TypeError: Unsupported entity type.
        """
        if isinstance(entity, OntologyIndividual):
            self._draw_node_individual(entity, complementary=complementary)
        elif isinstance(entity, OntologyRelationship):
            self._draw_node_relationship(entity, complementary=complementary)
        elif isinstance(entity, OntologyAttribute):
            self._draw_node_attribute(entity, complementary=complementary)
        elif isinstance(entity, OntologyClass):
            self._draw_node_class(entity, complementary=complementary)
        else:
            raise TypeError(f"Cannot draw {type(entity)}.")

    def _draw_node_individual(
        self,
        individual: OntologyIndividual,
        complementary: bool = False,
    ) -> None:
        """Add an ontology individual as a node to the graph.

        Args:
            individual: Ontology individual to draw.
            complementary: Whether the drawn entity is complementary or not.
        """
        attributes = self._attribute.format(
            "classes",
            ",".join(
                str(x) + (f" ({x.namespace.name})" if x.namespace else "")
                for x in individual.classes
            ),
        )

        for key, value in individual.attributes.items():
            label = self._get_element_label(key)
            if len(value) == 1:
                value = list(value)[0]
            elif len(value) == 0:
                value = None
            else:
                value = str(set(value)).replace(":", "_").replace("/", "_")
            attributes += self._attribute.format(label, str(value))

        if individual in self._requested_individuals:
            attributes += self._attribute.format(
                "session", self._get_element_label(individual.session)
            )
            extra_kwargs = {"color": "lightblue", "style": "filled"}
        else:
            extra_kwargs = dict()

        if complementary:
            extra_kwargs |= {
                "shape": "rectangle",
            }

        label = self._label.format(
            self._get_element_label(individual), attributes
        )

        self._graph.node(
            str(individual.identifier).replace(":", "_").replace("/", "_"),
            label=label,
            **extra_kwargs,
        )

    def _draw_node_class(
        self,
        class_: OntologyClass,
        complementary: bool = False,
    ) -> None:
        """Add an ontology class as a node to the graph.

        Args:
            class_: Ontology class to draw.
            complementary: Whether the drawn entity is complementary or not.
        """
        attr = ""
        for key, value in class_.attributes.items():
            attr += self._attribute.format(
                self._get_element_label(key),
                value if value else None,
            )
        label = (
            f"{self._get_element_label(class_)} "
            f'({class_.namespace.name if class_.namespace else ""})'
        )
        label = self._label.format(label, attr)

        if complementary:
            extra_kwargs = {
                "color": "#EEE4DD",
                "shape": "rectangle",
            }
        else:
            extra_kwargs = {"color": "#EED5C6"}

        self._graph.node(
            str(class_.identifier).replace(":", "_").replace("/", "_"),
            style="filled",
            label=label,
            **extra_kwargs,
        )

    def _draw_node_relationship(
        self,
        rel: OntologyRelationship,
        complementary: bool = False,
    ) -> None:
        """Add an ontology relationship as a node to the graph.

        Args:
            rel: Ontology class to draw.
            complementary: Whether the drawn entity is complementary or not.
        """
        attr = ""
        label = (
            f"{self._get_element_label(rel)} "
            f'({rel.namespace.name if rel.namespace else ""})'
        )
        label = self._label.format(label, attr)

        if complementary:
            extra_kwargs = {
                "color": "#DCDBEB",
                "shape": "rectangle",
            }
        else:
            extra_kwargs = {"color": "#AFABEB"}

        self._graph.node(
            str(rel.identifier).replace(":", "_").replace("/", "_"),
            label=label,
            style="filled",
            **extra_kwargs,
        )

    def _draw_node_attribute(
        self,
        attribute: OntologyAttribute,
        complementary: bool = False,
    ) -> None:
        """Add an ontology attribute as a node to the graph.

        Args:
            attribute: Ontology attribute to draw.
            complementary: Whether the drawn entity is complementary or not.
        """
        try:
            datatype = attribute.datatype
        except NotImplementedError:
            datatype = "multiple"
        attr = self._attribute.format("datatype", datatype)
        label = (
            f"{self._get_element_label(attribute)} "
            f'({attribute.namespace.name if attribute.namespace else ""})'
        )
        label = self._label.format(label, attr)

        if complementary:
            extra_kwargs = {
                "color": "#ADB8AB",
                "shape": "rectangle",
            }
        else:
            extra_kwargs = {"color": "#7EB874"}

        self._graph.node(
            str(attribute.identifier).replace(":", "_").replace("/", "_"),
            label=label,
            style="filled",
            **extra_kwargs,
        )

    def _draw_edge(
        self, start: OntologyEntity, end: OntologyEntity, **kwargs
    ) -> None:
        """Add an edge between two nodes.

        Args:
            start: start node
            end: end node
        """
        self._graph.edge(
            str(start.identifier).replace(":", "_").replace("/", "_"),
            str(end.identifier).replace(":", "_").replace("/", "_"),
            **kwargs,
        )


def terminal():
    """Run Semantic2Dot from the terminal."""
    # Parse the user arguments
    parser = argparse.ArgumentParser(description="Plot ontology namespaces.")
    parser.add_argument(
        "plot",
        metavar="plot",
        type=str,
        nargs="+",
        help="Either installed namespaces or paths to yaml ontology files",
    )
    parser.add_argument(
        "--output",
        "-o",
        type=os.path.abspath,
        default=Path(os.getcwd()) / "semantic2dot",
        help="The path of the output file",
    )
    args = parser.parse_args()

    namespaces = set()
    for x in args.plot:

        try:
            namespaces.add(Session.default_ontology.get_namespace(x))
            logger.warning("Using installed version of namespace %s" % x)
            continue
        except KeyError:
            pass

        parser = OntologyParser.get_parser(x)
        logger.warning("Including all namespaces of package %s" % x)
        for iri in parser.namespaces.values():
            try:
                namespace = Session.default_ontology.get_namespace(iri)
            except KeyError:
                Session.default_ontology.load_parser(parser)
                namespace = Session.default_ontology.get_namespace(iri)
            namespaces.add(namespace)

    # Convert the ontology to dot
    converter = Semantic2Dot(*namespaces)
    converter.render(filename=args.output_filename)


def semantic2dot(
    *elements: Union[
        OntologyIndividual,
        OntologyNamespace,
        Session,
    ],
    rel: Optional[
        Union[OntologyRelationship, Iterable[OntologyRelationship]]
    ] = None,
) -> Semantic2Dot:
    """Utility for plotting ontologies.

    Plot A-Boxes (ontology individuals) and the relationships between them,
    plot T-Boxes (classes, relationships and attributes), or a combination
    of them.

    Args:
        elements: Elements to plot:
            (Session) plot the whole contents of a session;
            (OntologyNamespace) plot all the ontology entities contained
                in the ontology namespace;
            (OntologyIndividual) plots an ontology individual,
                or a collection of them, and the relationships between them if
                multiple are provided;
        rel: When not `None` and when plotting an ontology individual, calls
            uses the method `find(individual, rel=rel, find_all=True)` from
            `simphony_osp.tools` to additionally plot such individuals.
    """
    return Semantic2Dot(*elements, rel=rel)


if __name__ == "__main__":
    terminal()



"""An annotation property defined in the ontology."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Iterable, Iterator, Optional

from rdflib import OWL, RDF, RDFS
from rdflib.term import Identifier

from simphony_osp.ontology.entity import OntologyEntity
from simphony_osp.utils.datatypes import UID, Triple

if TYPE_CHECKING:
    from simphony_osp.session.session import Session

logger = logging.getLogger(__name__)


class OntologyAnnotation(OntologyEntity):
    """An annotation property defined in the ontology."""

    rdf_type = {OWL.AnnotationProperty, RDF.Property}
    rdf_identifier = Identifier

    # Public API
    # ↓ ------ ↓

    # No methods meant for the user.

    # ↑ ------ ↑
    # Public API

    def __init__(
        self,
        uid: UID,
        session: Optional[Session] = None,
        triples: Optional[Iterable[Triple]] = None,
        merge: bool = False,
    ) -> None:
        """Initialize the ontology annotation.

        Args:
            uid: UID identifying the entity.
            session: Session where the entity is stored.
            triples: Construct the annotation with the provided triples.
            merge: Whether overwrite the potentially existing entity in the
                session with the provided triples or just merge them with
                the existing ones.
        """
        super().__init__(uid, session, triples, merge=merge)
        logger.debug("Create ontology annotation property %s" % self)

    def _get_direct_superclasses(self) -> Iterator[OntologyEntity]:
        """Get all the direct superclasses of this annotation.

        Returns:
            The direct superannotations.
        """
        return (
            self.session.from_identifier(o)
            for o in self.session.graph.objects(self.iri, RDFS.subPropertyOf)
        )

    def _get_direct_subclasses(self) -> Iterator[OntologyEntity]:
        """Get all the direct subclasses of this annotation.

        Returns:
            OntologyRelationship: The direct subannotations
        """
        return (
            self.session.from_identifier(s)
            for s in self.session.graph.subjects(RDFS.subPropertyOf, self.iri)
        )

    def _get_superclasses(self) -> Iterator[OntologyEntity]:
        """Get all the superclasses of this annotation.

        Yields:
            The superannotations.
        """
        yield self

        def closure(node, graph):
            yield from graph.objects(node, RDFS.subPropertyOf)

        yield from (
            self.session.from_identifier(x)
            for x in self.session.graph.transitiveClosure(
                closure, self.identifier
            )
        )

    def _get_subclasses(self) -> Iterator[OntologyEntity]:
        """Get all the subclasses of this annotation.

        Yields:
            The subannotations.
        """
        yield self

        def closure(node, graph):
            yield from graph.subjects(RDFS.subPropertyOf, node)

        yield from (
            self.session.from_identifier(x)
            for x in self.session.graph.transitiveClosure(
                closure, self.identifier
            )
        )



"""Interface between the SimPhoNy OSP and SQLite."""

from pathlib import Path

from simphony_osp.interfaces.sqlalchemy.interface import SQLAlchemy


class SQLite(SQLAlchemy):
    """An interface to an SQLite database.

    It just uses the SQLAlchemy wrapper with "sqlite" as prefix, so that the
    user does not need to care about it.
    """

    # TriplestoreInterface
    # ↓ ---------------- ↓

    def open(self, configuration: str, create: bool = False):
        """Open a connection to the database.

        Args:
            configuration: The path pointing to the file containing the
                SQLite database.
            create: Whether to create the database file if it does not exist.
        """
        if not create and not Path(configuration).is_file():
            raise FileNotFoundError(
                f"Database file {configuration} does not " f"exist."
            )
        return super().open(configuration="sqlite:///" + configuration)

    # ↑ ---------------- ↑



"""Implementation of a remote RDFLib store over websockets."""

import io
import json
import logging
import tempfile
from typing import BinaryIO, Callable, Dict, List, Tuple
from uuid import UUID

from rdflib import Graph, URIRef
from rdflib.plugins.parsers.jsonld import to_rdf as json_to_rdf

from simphony_osp.interfaces.interface import Interface
from simphony_osp.interfaces.remote.common import COMMAND
from simphony_osp.interfaces.remote.engine import CommunicationEngineServer

logger = logging.getLogger(__name__)


class InterfaceServer:
    """Receives commands from a client to drive an interface."""

    def __init__(
        self,
        host: str,
        port: int,
        generate_interface: Callable[[str, str], Interface],
    ):
        """Initialize the server."""
        self._engine: CommunicationEngineServer = CommunicationEngineServer(
            host=host,
            port=port,
            handle_request=self.handle_request,
            handle_disconnect=self.handle_disconnect,
        )
        self._interfaces: Dict[UUID, Interface] = dict()
        self._directories: Dict[UUID, tempfile.TemporaryDirectory] = dict()
        self._interface_generator: Callable[
            [str, str], Interface
        ] = generate_interface

    def listen(self) -> None:
        """Listen for connections from clients."""
        self._engine.listen()

    def handle_disconnect(self, connection_id: UUID) -> None:
        """Handle the disconnect of a user. Close and delete his session.

        Args:
            connection_id: The connection that has disconnected.
        """
        if connection_id in self._interfaces:
            self._interfaces[connection_id].close()
            del self._interfaces[connection_id]
        else:
            logger.warning(
                "User %s disconnected, even though it was not "
                "associated with a session." % connection_id
            )

    def handle_request(
        self,
        command: COMMAND,
        data: str,
        files: List[BinaryIO],
        connection_id: UUID,
    ) -> Tuple[str, list]:
        """Handle requests from the client."""
        try:
            if command == COMMAND.OPEN:
                response = self._open(data, connection_id)
            elif command == COMMAND.CLOSE:
                response = self._close(data, connection_id)
            elif command == COMMAND.POPULATE:
                response = self._populate(data, connection_id)
            elif command == COMMAND.COMMIT:
                response = self._commit(data, connection_id)
            elif command == COMMAND.COMPUTE:
                response = self._compute(data, connection_id)
            elif command == COMMAND.ADD:
                response = self._add(data, connection_id)
            elif command == COMMAND.TRIPLES:
                response = self._triples(data, connection_id)
            elif command == COMMAND.REMOVE:
                response = self._remove(data, connection_id)
            elif command == COMMAND.SAVE:
                response = self._save(data, files, connection_id)
            elif command == COMMAND.LOAD:
                response = self._load(data, connection_id)
            elif command == COMMAND.DELETE:
                response = self._delete(data, connection_id)
            elif command == COMMAND.HASH:
                response = self._hash(data, connection_id)
            elif command == COMMAND.RENAME:
                response = self._rename(data, connection_id)
            elif command == COMMAND.STORE_OPEN:
                response = self._store_open(data, connection_id)
            elif command == COMMAND.STORE_CLOSE:
                response = self._store_close(data, connection_id)
            elif command == COMMAND.STORE_ADD:
                response = self._store_add(data, connection_id)
            elif command == COMMAND.STORE_TRIPLES:
                response = self._store_triples(data, connection_id)
            elif command == COMMAND.STORE_REMOVE:
                response = self._store_remove(data, connection_id)
            elif command == COMMAND.STORE_COMMIT:
                response = self._store_commit(data, connection_id)
            elif command == COMMAND.HASATTR:
                response = self._hasattr(data, connection_id)
            elif command == COMMAND.AUTHENTICATE:
                response = self._authenticate(data, connection_id)
            else:
                response = "ERROR: Invalid command", []
            if isinstance(response, str):
                response = response, []
            return response
        except Exception as e:
            logger.error(str(e))
            return ("ERROR: %s: %s" % (type(e).__name__, e)), []

    # Commands

    def _open(self, data: str, connection_id: UUID) -> str:
        interface = self._interfaces[connection_id]
        data = json.loads(data)
        interface.open(**data)
        return json.dumps({COMMAND.OPEN: None})

    def _close(self, data: str, connection_id: UUID) -> str:
        interface = self._interfaces[connection_id]
        interface.close()
        return json.dumps({COMMAND.CLOSE: None})

    def _populate(self, data: str, connection_id: UUID) -> str:
        interface = self._interfaces[connection_id]
        interface.populate()
        return json.dumps({COMMAND.POPULATE: None})

    def _commit(self, data: str, connection_id: UUID) -> str:
        interface = self._interfaces[connection_id]
        interface.commit()
        return json.dumps({COMMAND.COMMIT: None})

    def _compute(self, data: str, connection_id: UUID) -> str:
        interface = self._interfaces[connection_id]
        data = json.loads(data)
        kwargs = data["kwargs"]
        interface.compute(**kwargs)
        return json.dumps({COMMAND.COMPUTE: None})

    def _add(self, data: str, connection_id: UUID) -> str:
        interface = self._interfaces[connection_id]
        graph = json_to_rdf(json.loads(data), Graph())
        for triple in graph:
            interface.add(triple)
        return json.dumps({COMMAND.ADD: None})

    def _remove(self, data: str, connection_id: UUID) -> str:
        interface = self._interfaces[connection_id]
        pattern = next(
            tuple(x if x != URIRef("none:None") else None for x in triple)
            for triple in Graph().parse(io.StringIO(data), format="turtle")
        )
        graph = Graph()
        graph.addN((s, p, o, graph) for s, p, o in interface.remove(pattern))
        return (
            f"{{"
            f'"{COMMAND.REMOVE}": '
            f'{graph.serialize(format="json-ld")}'
            f"}}"
        )

    def _triples(self, data: str, connection_id: UUID) -> str:
        interface = self._interfaces[connection_id]
        pattern = next(
            tuple(x if x != URIRef("none:None") else None for x in triple)
            for triple in Graph().parse(io.StringIO(data), format="turtle")
        )
        graph = Graph()
        graph.addN((s, p, o, graph) for s, p, o in interface.triples(pattern))
        return (
            f"{{"
            f'"{COMMAND.TRIPLES}": '
            f'{graph.serialize(format="json-ld")}'
            f"}}"
        )

    def _save(
        self, data: str, files: List[BinaryIO], connection_id: UUID
    ) -> str:
        interface = self._interfaces[connection_id]
        if hasattr(interface, "save"):
            data = json.loads(data)
            key = data["key"]
            file = files[0]
            interface.save(key, file)
            response = {COMMAND.SAVE: None}
        else:
            response = {COMMAND.NOT_FOUND: True}
        return json.dumps(response)

    def _load(self, data: str, connection_id: UUID) -> (str, List[BinaryIO]):
        interface = self._interfaces[connection_id]
        if hasattr(interface, "load"):
            data = json.loads(data)
            key = data["key"]
            file = interface.load(key)
            response = {COMMAND.LOAD: None}
            return json.dumps(response), [file]
        else:
            response = {COMMAND.NOT_FOUND: True}
            return json.dumps(response), []

    def _delete(self, data: str, connection_id: UUID) -> str:
        interface = self._interfaces[connection_id]
        if hasattr(interface, "delete"):
            data = json.loads(data)
            key = data["key"]
            interface.delete(key)
            response = {COMMAND.DELETE: None}
        else:
            response = {COMMAND.NOT_FOUND: True}
        return json.dumps(response)

    def _hash(self, data: str, connection_id: UUID) -> str:
        interface = self._interfaces[connection_id]
        if hasattr(interface, "delete"):
            data = json.loads(data)
            key = data["key"]
            file_hash = interface.hash(key)
            response = {COMMAND.HASH: file_hash}
        else:
            response = {COMMAND.NOT_FOUND: True}
        return json.dumps(response)

    def _rename(self, data: str, connection_id: UUID) -> str:
        interface = self._interfaces[connection_id]
        if hasattr(interface, "rename"):
            data = json.loads(data)
            key = data["key"]
            new_key = data["new_key"]
            interface.rename(key, new_key)
            response = {COMMAND.RENAME: None}
        else:
            response = {COMMAND.NOT_FOUND: True}
        return json.dumps(response)

    def _store_open(self, data: str, connection_id: UUID) -> str:
        interface = self._interfaces[connection_id]
        data = json.loads(data)
        interface.base.open(**data)
        return json.dumps({COMMAND.STORE_OPEN: None})

    def _store_close(self, data: str, connection_id: UUID) -> str:
        interface = self._interfaces[connection_id]
        data = json.loads(data)
        interface.base.close(**data)
        return json.dumps({COMMAND.STORE_CLOSE: None})

    def _store_add(self, data: str, connection_id: UUID) -> str:
        interface = self._interfaces[connection_id]
        graph = json_to_rdf(json.loads(data), Graph())
        interface.base.addN((s, p, o, interface.base) for s, p, o in graph)
        return json.dumps({COMMAND.STORE_ADD: None})

    def _store_remove(self, data: str, connection_id: UUID) -> str:
        interface = self._interfaces[connection_id]
        patterns = (
            tuple(x if x != URIRef("none:None") else None for x in triple)
            for triple in Graph().parse(io.StringIO(data), format="turtle")
        )
        for pattern in patterns:
            interface.base.remove(pattern)
        return json.dumps({COMMAND.REMOVE: None})

    def _store_triples(self, data: str, connection_id: UUID) -> str:
        interface = self._interfaces[connection_id]
        pattern = next(
            tuple(x if x != URIRef("none:None") else None for x in triple)
            for triple in Graph().parse(io.StringIO(data), format="turtle")
        )
        graph = Graph()
        graph.addN(
            (s, p, o, graph) for s, p, o in interface.base.triples(pattern)
        )
        return (
            f"{{"
            f'"{COMMAND.STORE_TRIPLES}": '
            f'{graph.serialize(format="json-ld")}'
            f"}}"
        )

    def _store_commit(self, data: str, connection_id: UUID) -> str:
        interface = self._interfaces[connection_id]
        interface.base.commit()
        return json.dumps({COMMAND.STORE_COMMIT: None})

    def _hasattr(self, data: str, connection_id: UUID) -> str:
        interface = self._interfaces[connection_id]
        data = json.loads(data)
        attr = data["item"]
        return json.dumps({COMMAND.HASATTR: hasattr(interface, attr)})

    def _authenticate(self, data: str, connection_id: UUID) -> str:
        if connection_id not in self._interfaces:
            data = json.loads(data)
            username, password = data["username"], data["password"]
            self._interfaces[connection_id] = self._interface_generator(
                username, password
            )
        return json.dumps({COMMAND.AUTHENTICATE: None})



"""You can import the installed namespaces from this module."""
from __future__ import annotations

import logging as _logging
from typing import TYPE_CHECKING as _TYPE_CHECKING
from typing import Union as _Union

from rdflib import URIRef as _URIRef
from rdflib.term import Identifier as _Identifier

from simphony_osp.session.session import Session as _Session

if _TYPE_CHECKING:
    from simphony_osp.ontology.entity import OntologyEntity

self = __import__(__name__)

_logger = _logging.getLogger(__name__)


# Access namespaces from this module.


def __getattr__(name: str):
    try:
        return _Session.default_ontology.get_namespace(name)
    except KeyError as e:
        raise AttributeError from e


def __dir__():
    return list(x.name for x in _Session.default_ontology.namespaces)


__all__ = __dir__()


# `from_iri` as gateway to `_tbox.from_identifier`.
def from_iri(iri: _Union[str, _URIRef]):
    """Get an entity from its IRI from the default TBox.

    Args:
        iri: The IRI of the entity.

    Raises:
        KeyError: The ontology entity is not stored in the default TBox.

    Returns:
        The OntologyEntity.
    """
    if type(iri) is str:
        iri = _URIRef(iri)
    if not isinstance(iri, _URIRef):
        raise TypeError(f"Expected {str} or {_URIRef}, not {type(iri)}.")
    return _Session.default_ontology.from_identifier(iri)


# `from_identifier` as gateway to `_tbox.from_identifier`.
def from_identifier(identifier: _Identifier) -> OntologyEntity:
    """Get an entity from its identifier from the default TBox.

    Args:
        identifier: The identifier of the entity.

    Raises:
        KeyError: The ontology entity is not stored in the default TBox.

    Returns:
        The OntologyEntity.
    """
    return _Session.default_ontology.from_identifier(identifier)



"""A collection of utility methods for SimPhoNy.

These are potentially useful for every user of SimPhoNy.
"""

import logging
from typing import Set

from simphony_osp.ontology.individual import OntologyIndividual
from simphony_osp.ontology.relationship import OntologyRelationship

logger = logging.getLogger(__name__)

__all__ = [
    "branch",
    "relationships_between",
]


def branch(
    individual, *individuals, rel: OntologyRelationship
) -> OntologyIndividual:
    """Like `connect`, but returns the element you connect to.

    This makes it easier to create large structures involving ontology
    individuals.

    Args:
        individual: The ontology individual that is the subject of the
            connections to be created.
        individuals: Ontology individuals to connect to.
        rel: Relationship to use.

    Returns:
        The ontology individual that is the subject of the connections to be
        created (the first argument).
    """
    individual.connect(*individuals, rel=rel)
    return individual


def relationships_between(
    subj: OntologyIndividual, obj: OntologyIndividual
) -> Set[OntologyRelationship]:
    """Get the set of relationships between two cuds objects.

    Args:
        subj: The subject ontology individual.
        obj: The object ontology individual.

    Returns:
        The set of relationships between subject and object.
    """
    return {
        relationship
        for individual, relationship in subj.get(obj, return_rel=True)
    }



"""Utilities used in the communication for the remote stores."""

import asyncio
import logging
import math
import os
import tempfile
import uuid
from itertools import chain
from typing import (
    Any,
    BinaryIO,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)
from uuid import UUID

import websockets
import websockets.exceptions as ws_exceptions
from websockets.legacy.client import WebSocketClientProtocol as Socket
from websockets.legacy.server import WebSocketServerProtocol as ServerSocket

from simphony_osp.interfaces.remote.common import COMMAND

logger = logging.getLogger(__name__)

BLOCK_SIZE = 4096
LEN_FILES_HEADER = [5]  # num_blocks
LEN_HEADER = [2, 5, 2]  # version, num_blocks, num_files
VERSION = 2
DEBUG_MAX = 1000


def clear_directory(directory: str) -> None:
    """Remove all the files and subdirectories contained within a folder."""
    for path, directory_names, file_names in os.walk(directory):
        for element in chain(file_names, directory_names):
            os.unlink(os.path.join(path, element))


def decode_header(bytestring: bytes, lengths: List[int]) -> Union[int, str]:
    """Decode the header given as a string of bytes.

    Args:
        bytestring: The encoded header
        lengths: The number of bytes for the individual components

    Yields:
        Interpret all elements that have a corresponding length as int. If
            there are leftover bytes decode them using utf-8.
    """
    i = 0
    for length in lengths:
        if i + length > len(bytestring):
            raise IndexError("Length mismatch in header")
        yield int.from_bytes(bytestring[i : i + length], byteorder="big")
        i += length
    if len(bytestring) > i:
        yield bytestring[i:].decode("utf-8")


def encode_header(
    elements: List[Union[int, str]], lengths: List[int]
) -> bytes:
    """Encode the header to single array of bytes.

    Args:
        elements: The elements to encode. All but the last
                                    must be int. The last one can be str.
        lengths ([type]): The number of bytes in the encoded header.

    Raises:
        ValueError: elements and lengths mismatch in number of elements
        NotImplementedError: Invalid datatype in elements.

    Returns:
        bytes: The encoded header
    """
    r = b""
    if len(lengths) > len(elements) or len(elements) > len(lengths) + 1:
        raise ValueError
    for element, length in zip(elements, lengths):
        if not isinstance(element, int):
            raise NotImplementedError
        r += element.to_bytes(length=length, byteorder="big")
    if len(elements) > len(lengths):
        if not isinstance(elements[-1], str):
            raise NotImplementedError("Invalid type of %s" % elements[-1])
        r += elements[-1].encode("utf-8")
    return r


def split_message(msg: str, block_size: int = BLOCK_SIZE):
    """Split the message to send in small blocks.

    Args:
        msg: The message to send.
        block_size: The size of the blocks.
                                    Defaults to BLOCK_SIZE.

    Returns:
        int, Generator: Number of blocks, Generator over blocks.
    """
    msg = msg.encode("utf-8")
    num_blocks = int(math.ceil(len(msg) / block_size))

    def gen(message, blocks, size):
        for i in range(blocks):
            logger.debug(f"Sending message block {i + 1} of {blocks}")
            yield message[i * size : (i + 1) * size]
        logger.debug("Done")

    return num_blocks, gen(msg, num_blocks, block_size)


async def join_message(websocket, num_blocks: int) -> str:
    """Get the message that was decomposed in different blocks.

    Args:
        websocket: wecksocket object to receive the objects.
        num_blocks: The number of blocks that belong to the message.

    Returns:
        str: The data, decoded using utf-8.
    """
    data = b""
    for i in range(num_blocks):
        logger.debug(f"Receiving message block {i + 1} of {num_blocks}")
        data += await websocket.recv()
    logger.debug("Done")
    data = data.decode("utf-8")
    return data


def encode_files(files: List[str]) -> bytes:
    """Encode the files to be sent to over the networks.

    Will send file in several blocks.

    Args:
        files: A list of paths to send.

    Yields:
        bytes: The bytes of the file
    """
    logger.debug("Will send %s files" % len(files))
    for i, file in enumerate(files):
        filename = os.path.basename(file)
        num_blocks = int(math.ceil(os.path.getsize(file) / BLOCK_SIZE))
        logger.debug(
            "Send file %s (%s of %s) with %s block(s) of %s bytes"
            % (file, i + 1, len(files), num_blocks, BLOCK_SIZE)
        )
        yield encode_header([num_blocks, filename], LEN_FILES_HEADER)

        # send the file contents
        with open(file, "rb") as f:
            for i, block in enumerate(iter(lambda: f.read(BLOCK_SIZE), b"")):
                logger.debug(f"Send file block {i + 1} of {num_blocks}")
                yield block
            logger.debug("Done")


async def receive_files(
    num_files: int, websocket: Union[Socket, ServerSocket]
) -> List[BinaryIO]:
    """Will receive and store the files sent to the websocket.

    Args:
        num_files: The number of files to load.
        websocket: The websocket to load the files from.
    """
    files = []
    for i in range(num_files):
        logger.debug(f"Load file {i + 1} of {num_files}")
        num_blocks, filename = decode_header(
            await websocket.recv(), LEN_FILES_HEADER
        )
        file = tempfile.NamedTemporaryFile(delete=False)
        files.append(file)
        logger.debug("Storing file %s with %s blocks." % file.name)
        with file:
            for j in range(num_blocks):
                logger.debug(f"Receive block {j + 1} of {num_blocks}")
                data = await websocket.recv()
                file.write(data)
        file.seek(0)
    return files


class CommunicationEngineServer:
    """Server side of the CommunicationEngine.

    The communication engine manages the connection between the remote and
    a local side. The server will be executed on the remote side.
    """

    host: str
    port: int
    credentials: Dict[UUID, Tuple[str, str]]

    def __init__(
        self,
        host: str,
        port: int,
        handle_request: Callable[
            [COMMAND, str, List[BinaryIO], UUID], Tuple[str, List[str]]
        ],
        handle_disconnect: Callable[[UUID], None],
    ) -> None:
        """Construct the communication engine's server.

        Args:
            host: The hostname.
            port: The port.
            handle_request: Handles the requests of the user.
            handle_disconnect: Gets called when a user disconnects.
        """
        self.host = host
        self.port = port
        self._handle_request = handle_request
        self._handle_disconnect = handle_disconnect
        self._connections = dict()
        self.credentials = dict()

    def listen(self) -> None:
        """Start the server on given host and port."""
        event_loop = asyncio.get_event_loop()
        start_server = websockets.serve(self._serve, self.host, self.port)
        event_loop.run_until_complete(start_server)
        event_loop.run_forever()

    async def _serve(self, socket: ServerSocket, _: str) -> None:
        """Wait for requests, compute responses and serve them to the user.

        Args:
            socket: The websockets object.
        """
        self._connections[socket] = self._connections.get(socket, uuid.uuid4())
        connection = self._connections[socket]

        try:
            while True:
                # receive the request
                command, data, files = await self._decode(socket)
                command = COMMAND(command)  # Validate command
                # handle the request and produce a response
                response, response_files = self._handle_request(
                    command, data, files, connection
                )
                # clean-up received files
                for file in files:
                    file.close()
                    os.remove(file.name)
                # send the response
                logger.debug(
                    "Response: %s with %s files"
                    % (response[:DEBUG_MAX], len(response_files))
                )
                num_blocks, response = split_message(response)
                await socket.send(
                    encode_header(
                        [VERSION, num_blocks, len(response_files)], LEN_HEADER
                    )
                )
                for part in response:
                    await socket.send(part)
                # send response files
                if len(response_files) > 0:
                    with tempfile.TemporaryDirectory() as temp_dir:
                        file_names = []
                        for i, file in enumerate(response_files):
                            file_name = os.path.join(temp_dir.name, str(i))
                            with open("wb", file_name) as tmp_file:
                                tmp_file.write(file)
                            file_names.append(file_name)
                        for part in encode_files(file_names):
                            await socket.send(part)
        except ws_exceptions.ConnectionClosedOK:
            pass
        finally:
            logger.debug("Connection %s closed!" % connection)
            del self._connections[socket]
            self._handle_disconnect(connection)

    async def _decode(
        self, socket: ServerSocket
    ) -> Tuple[str, str, List[BinaryIO]]:
        """Get data from the user.

        Args:
            socket: The websocket object

        Raises:
            NotImplementedError: Not implemented the protocol version of
                the message. You might need to update SimPhoNy.

        Returns:
            Tuple of command to execute and data
        """
        connection = self._connections[socket]

        bytes_data = await socket.recv()
        version, num_blocks, num_files, command = decode_header(
            bytes_data, LEN_HEADER
        )
        if version != VERSION:
            raise NotImplementedError(
                "No decode implemented for " "version %s" % version
            )
        logger.debug(
            "Received data from %s.\n\t Protocol version: %s,\n\t "
            "Command: %s,\n\t Number of files: %s."
            % (connection, version, command, num_files)
        )
        data = await join_message(socket, num_blocks)
        logger.debug("Received data: %s" % data[:DEBUG_MAX])
        files = await receive_files(num_files, socket)
        return command, data, files


class CommunicationEngineClient:
    """Client side of the CommunicationEngine.

    The communication engine manages the connection between a remote and
    local side. The client will be executed on the local side.
    """

    socket: Optional[Socket] = None

    def __init__(self, uri: str, handle_response: Callable[..., Any]):
        """Construct the communication engine's client.

        Args:
            uri: WebSocket URI.
            handle_response: Handles the responses of the server.
                Signature: str(response).
        """
        self.uri = uri
        self._handle_response = handle_response

    def send(
        self,
        command: COMMAND,
        data: str,
        files: Optional[Iterable[str]] = None,
    ):
        """Send a request to the server.

        Args:
            command: The command to execute on the server.
            data: The data to send to the server.
            files: List of file paths.

        Returns:
            Future: The Future’s result or raise its exception.
        """
        files = files or []
        event_loop = asyncio.get_event_loop()
        return event_loop.run_until_complete(
            self._request(command, data, files)
        )

    def close(self) -> None:
        """Close the connection to the server."""
        event_loop = asyncio.get_event_loop()
        event_loop.run_until_complete(self._close())

    async def _request(
        self, command: COMMAND, data: str, files: List[str]
    ) -> str:
        """Send a request to the server.

        Args:
            command: The command to execute on the server.
            data: The data to send to the server.

        Returns:
            The response for the client.
        """
        logger.debug(f"Request {command}: {data[:DEBUG_MAX]}")
        if self.socket is None:
            logger.debug("uri: %s" % self.uri)
            self.socket = await websockets.connect(self.uri)

        # send request to the server
        message = self._encode(command, data, files)
        for part in message:
            await self.socket.send(part)

        # handle response
        response = await self.socket.recv()
        version, num_blocks, num_files = decode_header(response, LEN_HEADER)
        logger.debug(
            "Response:\n\t Protocol version: %s,\n\t "
            "Number of blocks: %s,\n\t Number of files: %s"
            % (version, num_blocks, num_files)
        )
        data = await join_message(self.socket, num_blocks)
        logger.debug("Response data: %s" % data[:DEBUG_MAX])
        files = await receive_files(num_files, self.socket)
        return self._handle_response(data=data, files=files)

    async def _close(self):
        """Close the connection to the server."""
        if self.socket is not None:
            await self.socket.close()
            self.socket = None

    def __del__(self):
        """Close the connection on garbage collection."""
        self.close()

    @staticmethod
    def _encode(command: COMMAND, data: str, files: List[str]) -> bytes:
        """Encode the data to send to the server to bytes.

        Args:
            command: The command to execute.
            data: The json data to send.
            files: The binary data to send.

        Returns:
            bytes: The resulting data encoded
        """
        files = files
        num_blocks, data = split_message(data)
        version = VERSION
        yield encode_header(
            [version, num_blocks, len(files), command], LEN_HEADER
        )
        yield from data
        yield from encode_files(files)



"""This module contains methods for datatype conversions."""
from __future__ import annotations

import logging
import re
import sys
from abc import ABC
from base64 import b85decode, b85encode
from datetime import datetime
from decimal import Decimal
from distutils.version import StrictVersion
from fractions import Fraction
from typing import (
    TYPE_CHECKING,
    Any,
    List,
    MutableSequence,
    Optional,
    Sequence,
    Tuple,
    Union,
)
from uuid import UUID, uuid4

import numpy as np
import rdflib
from rdflib import RDFS, XSD, BNode, Literal, URIRef, term
from rdflib.term import Identifier, Node

if TYPE_CHECKING:
    from simphony_osp.ontology.annotation import OntologyAnnotation
    from simphony_osp.ontology.attribute import OntologyAttribute
    from simphony_osp.ontology.individual import OntologyIndividual
    from simphony_osp.ontology.oclass import OntologyClass
    from simphony_osp.ontology.relationship import OntologyRelationship

# Some IDEs may not detect a type alias defined as Union[...] otherwise.
python_version_supports_type_alias = (
    StrictVersion("3.10")
    <= StrictVersion(f"{sys.version_info.major}.{sys.version_info.minor}")
    < StrictVersion("4.0")
)
if python_version_supports_type_alias:
    from typing import TypeAlias
else:
    from typing import Any as TypeAlias

__all__ = [
    "ATTRIBUTE_VALUE_TYPES",
    "AttributeValue",
    "AnnotationValue",
    "Identifier",
    "NestedList",
    "NestedMutableSequence",
    "NestedSequence",
    "NestedTuple",
    "OntologyPredicate",
    "Pattern",
    "PredicateValue",
    "RelationshipValue",
    "SimplePattern",
    "SimpleTriple",
    "Triple",
    "UID",
    "Vector",
    "rdf_to_python",
]

# Collection type hints
NestedSequence = Union[Sequence[Any], Sequence["NestedSequence"]]
NestedMutableSequence = Union[
    MutableSequence[Any], MutableSequence["NestedSequence"]
]
NestedTuple = Union[Tuple[Any, ...], Tuple["NestedTuple", ...]]
NestedList = Union[List[Any], List["NestedList"]]


# Ignore RDFLib binding warning.
class _BindingWarningFilter(logging.Filter):
    """Filters datatype rebinding warnings from RDFLib."""

    def filter(self, record):
        """Checks that the string is not contained in the message."""
        return " was already bound. Rebinding." not in record.getMessage()


rdflib_term_logger = logging.getLogger(rdflib.term.__name__)
rdflib_term_logger.addFilter(_BindingWarningFilter())


class CustomDataType(ABC):
    """Abstract class for the Python representation of custom data types.

    Such custom data types must also have an RDF representation (the IRI
    attribute and a string representation).

    Beware that any custom datatype will need to have a "lexicalizer" and a
    "constructor" for rdflib compatibility. See the piece of code
    immediately after this class for more details.
    """

    iri: URIRef


"""After the custom data type has been defined, the rdflib lexicalizer
and constructor have to be specified.

The rdflib lexicalizer creates a string representation of the Python object.
For example, for vectors it is just a string representation of their binary
representation. This is string representation is what the user will find in
exported json or rdf files, and also what literals print to the Python console
through the __repr__ method.

The rdflib constructor does the opposite, create a Python representation
from a literal constructed from a string. For example, for vectors,
it converts the string representation of their binary form into the Vector
python object.

Both the lexicalizer and the constructor might be optional. For more
information, please check  the documentation and source code for
[`rdflib.term.bind`]
(https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html#rdflib.term.bind).
"""


class Vector(CustomDataType):
    """Custom data type representing vectors.

    The data type is based on an underlying numpy array, or more
    specifically, the binary data for its shape, datatype and content.

    The reason behind not using directly a numpy array is that such objects
    are not hashable, but an RDF graph is a _set_ of triples, so having
    non-hashable objects leads to problems when it comes to handling and
    serializing the object.
    """

    # TODO: improve the interaction of the user with the object if possible:
    #   - If possible create the underlying numpy array when passing it as
    #     an argument to numpy's `array` function, instead of creating an array
    #     of vector objects.
    #   - Implement some operations with numpy arrays that return numpy arrays,
    #     just as it was done in the AttributeSet case.
    iri = URIRef("https://www.simphony-project.eu/types#Vector")

    _ELEMENT_LEN: int = 3
    _DTYPE_LEN: int = 1
    _SHAPE_LEN: int = 1

    @property
    def data(self) -> np.ndarray:
        """Returns the underlying numpy array, constructed from static data."""
        return np.frombuffer(self.__bytes, dtype=self.__dtype).reshape(
            self.__shape
        )

    def __getitem__(self, item: int):
        """Slice the underlying numpy array."""
        return self.data[item]

    def __init__(
        self, value: Union[Literal, np.ndarray, NestedSequence, Vector]
    ):
        """Creates a new vector from `value`.

        The process of creating such vector can be divided in several steps:
            1. Construct a numpy array from `value`.
            2. Read its `dtype` and `shape` and store them as private
               attributes.
            3. Convert to bytes the data contained in the array and store
               them in the `__bytes` private attribute.
        """
        if isinstance(value, Vector):
            array = np.array(value.data)
        else:
            array = np.array(value)
        # TODO: restrict datatypes.
        self.__dtype = array.dtype
        self.__shape = array.shape
        self.__bytes = array.tobytes()

    def __str__(self) -> str:
        """Convert the vector to a string."""
        return self.data.__str__()

    def __repr__(self) -> str:
        """String representation for the vector (i.e. on Python console)."""
        array_repr = self.data.__repr__()
        array_repr = array_repr.replace("\n", " ")
        array_repr = re.sub(r"\s{2,}", " ", array_repr)
        return (
            f"<{self.__class__.__module__}.{self.__class__.__name__}:"
            f" {array_repr}>"
        )

    def __hash__(self):
        """Get the vector's hash."""
        return hash((self.__bytes, self.__dtype, self.__shape))

    def __eq__(
        self, other: Union[Literal, np.ndarray, NestedSequence, Vector]
    ):
        """Compare the vector to another object that can be vectorized."""
        return np.array_equal(self.data, Vector(other).data)

    def to_blob(self) -> bytes:
        """Convert the vector to a bytes representation.

        Use `from_blob` to convert back to a vector.
        """
        blob = bytes()
        # Attach data type.
        blob += len(bytes(self.__dtype.str, encoding="UTF-8")).to_bytes(
            self._DTYPE_LEN, "big"
        )
        blob += bytes(self.__dtype.str, encoding="UTF-8")
        # Attach shape.
        blob += len(self.__shape).to_bytes(self._SHAPE_LEN, "big")
        for integer in self.__shape:
            blob += integer.to_bytes(self._ELEMENT_LEN, "big")
        # Attach array.
        blob += self.__bytes
        return blob

    @classmethod
    def from_blob(cls, blob: bytes) -> Vector:
        """Convert a bytes representation of a vector into a Vector object."""
        # Get data type.
        start_dtype = 0
        len_dtype = int.from_bytes(blob[start_dtype : cls._DTYPE_LEN], "big")
        dtype = np.dtype(
            blob[
                start_dtype
                + cls._DTYPE_LEN : start_dtype
                + cls._DTYPE_LEN
                + len_dtype
            ].decode()
        )
        # Get shape.
        start_shape = start_dtype + cls._DTYPE_LEN + len_dtype
        len_shape = int.from_bytes(
            blob[start_shape : start_shape + cls._SHAPE_LEN], "big"
        )
        shape = tuple(
            int.from_bytes(
                blob[
                    start_shape
                    + cls._SHAPE_LEN
                    + i * cls._ELEMENT_LEN : start_shape
                    + cls._SHAPE_LEN
                    + (i + 1) * cls._ELEMENT_LEN
                ],
                "big",
            )
            for i in range(0, len_shape)
        )
        # Get byte array.
        start_array = (
            start_shape + cls._SHAPE_LEN + len_shape * cls._ELEMENT_LEN
        )
        byte_array = blob[start_array:]
        array = np.frombuffer(byte_array, dtype=dtype).reshape(shape)
        return Vector(array)

    def to_b85(self) -> str:
        """Convert the vector to a base85 representation.

        Use `from_b85` to convert back to a vector.
        """
        return b85encode(self.to_blob()).decode("UTF-8")

    @classmethod
    def from_b85(cls, string) -> Vector:
        """Convert a base 85 representation of a vector back into a Vector."""
        return cls.from_blob(b85decode(string.encode("UTF-8")))

    @classmethod
    def from_blob_or_b85(cls, value: Union[str, bytes]) -> Vector:
        """Restore a vector from a base 85 or bytes representation of it."""
        if isinstance(value, bytes):
            return cls.from_blob(value)
        elif isinstance(value, str):
            return cls.from_b85(value)
        else:
            raise TypeError("Only str and bytes are allowed.")


for datatype in (Literal, np.ndarray, Sequence, list, tuple, Vector):
    term.bind(
        Vector.iri,
        datatype,
        lexicalizer=lambda x: Vector(x).to_b85(),
        constructor=lambda x: Vector.from_blob_or_b85(x),
        datatype_specific=True,
    )


class UID(CustomDataType):
    """Custom data type representing the unique identifier of an entity.

    This data type is actually not used in the RDF graph itself (an RDF
    identifier is used instead), but rather within SimPhoNy to identify the
    ontology entities objects. It plays the same role as a semantic web
    identifier, and can always be converted to one. In fact, such conversion
    is always performed before it is used in a graph operation. The only
    difference is that it can be based not only on semantic web identifiers
    such as blank nodes or IRIs, but also on other datatypes that can be
    converted on the fly to semantic-web compatible identifiers.

    However, if you want to store a UID as a literal for whatever reason,
    you are welcome to use this custom data type.
    """

    iri = URIRef("https://www.simphony-project.eu/types#UID")
    data: Union[Node, UUID]

    __slots__ = ("data",)

    def __init__(
        self,
        value: Optional[Union[UID, UUID, str, Node, int, bytes]] = None,
    ):
        """Initializes a new UID from `value`."""
        super().__init__()
        invalid = False
        if value is None:
            value = uuid4()
        elif isinstance(value, str):
            if value.startswith(ENTITY_IRI_PREFIX):
                value = value[len(ENTITY_IRI_PREFIX) :]
                try:
                    value = UUID(hex=value)
                except ValueError:
                    invalid = True
            elif not isinstance(value, Node):
                value = URIRef(value)
        elif isinstance(value, UUID):
            invalid = value.int == 0
        elif isinstance(value, int):
            value = UUID(int=value)
        elif isinstance(value, bytes):
            value = UUID(bytes=value)
        elif isinstance(value, UID):
            value = value.data
        if invalid or not isinstance(value, (UUID, Node)):
            raise ValueError(f"Invalid uid: {value}.")
        self.data = value

    def __hash__(self):
        """Get the uid's hash."""
        return hash(self.data)

    def __eq__(self, other):
        """Compare the uid to other uids.

        Returns false if `other` does not belong to the same class.
        """
        return isinstance(other, self.__class__) and self.data.__eq__(
            other.data
        )

    def __repr__(self):
        """String representation for the uid (i.e. on the Python console)."""
        return (
            f"<class '{self.__class__.__module__}"
            f".{self.__class__.__name__}':"
            f" {self.data.__repr__()}>"
        )

    def __str__(self):
        """Convert the uid to a string."""
        return self.data.__str__()

    def __format__(self, format_spec):
        """Formatter for UID objects."""
        return self.data.__format__(format_spec)

    def to_iri(self) -> URIRef:
        """Convert the UID to an IRI."""
        data = self.data
        if isinstance(data, UUID):
            return_iri = URIRef(ENTITY_IRI_PREFIX + str(data))
        elif isinstance(data, URIRef):
            return_iri = data
        elif isinstance(data, BNode):
            raise TypeError(
                f"The UID {self} cannot be converted to an IRI, "
                f"as it is a blank node."
            )
        else:
            raise RuntimeError(f"Illegal UID type {type(data)}.")
        return return_iri

    def to_uuid(self) -> UUID:
        """Convert the UID to a UUID.

        Will only work if the underlying data is a UUID.
        """
        data = self.data
        if not isinstance(data, UUID):
            raise Exception(
                f"Tried to get the UUID of the UID object "
                f"{self}, but this object is not a UUID."
            )
        return data

    def to_identifier(self) -> Identifier:
        """Convert the UID to a RDFLib Identifier."""
        data = self.data
        # logic in `to_iri` duplicated here in exchange for performance gains
        return (
            URIRef(ENTITY_IRI_PREFIX + str(data))
            if isinstance(data, UUID)
            else data
        )


for datatype in (UID, UUID, URIRef, str, int, bytes):
    if datatype in (int, bytes):

        def _lexicalizer(x):
            return str(UID(x))

    else:
        _lexicalizer = None
    term.bind(
        UID.iri,
        datatype,
        lexicalizer=_lexicalizer,
        constructor=None,
        datatype_specific=True,
    )

ENTITY_IRI_PREFIX = "https://www.simphony-project.eu/entity#"

# --- RDF TO PYTHON --- #


def _get_all_python_subclasses(cls):
    return set(cls.__subclasses__()) | {
        sc
        for c in cls.__subclasses__()
        for sc in _get_all_python_subclasses(c)
    }


# OWL data types
#  See [Datatype Maps](https://www.w3.org/TR/owl2-syntax/#Datatype_Maps) on
#  the OWL specification.
#  OWL_TO_PYTHON gathers data types that are not bound already to RDFLib terms.
OWL_TO_PYTHON = {
    URIRef("http://www.w3.org/2002/07/owl#real"): float,
    # ↑ Not defined by RDFLib.
    URIRef("http://www.w3.org/2002/07/owl#rational"): Fraction,
    # ↑ Not defined by RDFLib.
    # XSD.Name: ?, (TODO: not supported)
    # ↑ Not defined by RDFLib.
    # XSD.NCName: ?, (TODO: not yet supported)
    # ↑ Not defined by RDFLib.
    # XSD.NMTOKEN: ?, (TODO: not yet supported)
    # ↑ Not defined by RDFLib.
    # XSD.base64Binary: ?, (TODO: not yet supported)
    # ↑ Not defined by RDFLib.
    # XSD.dateTimeStamp: ?, (TODO: not yet supported)
    # ↑ ¿Is it defined by RDFLib?
    # RDF.XMLLiteral: ?, (TODO: not yet supported)
    # ↑ ¿Is it defined by RDFLib?
}
#  OWL_TO_PYTHON_DEFINED_RDFLIB gathers data types that were bound to RDFLib
#  terms already by the RDFLib developers.
OWL_TO_PYTHON_DEFINED_RDFLIB = {
    XSD.decimal: Decimal,  # -> Defined already by RDFLib.
    XSD.integer: int,  # -> Defined already by RDFLib.
    XSD.nonNegativeInteger: int,  # -> Defined already by RDFLib.
    XSD.nonPositiveInteger: int,  # -> Defined already by RDFLib.
    XSD.positiveInteger: int,  # -> Defined already by RDFLib.
    XSD.negativeInteger: int,  # -> Defined already by RDFLib.
    XSD.long: int,  # -> Defined already by RDFLib.
    XSD.int: int,  # -> Defined already by RDFLib.
    XSD.short: int,  # -> Defined already by RDFLib.
    XSD.byte: int,  # -> Defined already by RDFLib.
    XSD.unsignedLong: int,  # -> Defined already by RDFLib.
    XSD.unsignedInt: int,  # -> Defined already by RDFLib.
    XSD.unsignedShort: int,  # -> Defined already by RDFLib.
    XSD.unsignedByte: int,  # -> Defined already by RDFLib.
    XSD.double: float,  # -> Defined already by RDFLib.
    XSD.float: float,  # -> Defined already by RDFLib.
    XSD.string: str,  # -> Defined already by RDFLib.
    XSD.normalizedString: str,  # -> Defined already by RDFLib.
    XSD.token: str,  # -> Defined already by RDFLib.
    XSD.boolean: bool,  # -> Defined already by RDFLib.
    XSD.hexBinary: bytes,  # -> Defined already by RDFLib.
    XSD.anyURI: str,  # -> Defined already by RDFLib.
    XSD.dateTime: datetime,  # -> Defined already by RDFLib.
}
OWL_TO_PYTHON.update(OWL_TO_PYTHON_DEFINED_RDFLIB)
OWL_COMPATIBLE_TYPES = tuple(OWL_TO_PYTHON.values())
OWLCompatibleType = Union[OWL_COMPATIBLE_TYPES]
# Custom STATIC data types (fixed URI).
CUSTOM_TO_PYTHON = {
    cls.iri: cls for cls in _get_all_python_subclasses(CustomDataType)
}
CUSTOM_COMPATIBLE_TYPES = tuple(CUSTOM_TO_PYTHON.values())
CustomCompatibleType = (
    Union[CUSTOM_COMPATIBLE_TYPES] if CUSTOM_COMPATIBLE_TYPES else None
)
# All RDF data types compatible with SimPhoNy, automatically generated from
#  the previous mappings.
RDF_TO_PYTHON = {RDFS.Literal: str}
RDF_TO_PYTHON.update(OWL_TO_PYTHON)
RDF_TO_PYTHON.update(CUSTOM_TO_PYTHON)
#  Bind all the types to RDFLib.
for iri, data_type in RDF_TO_PYTHON.items():
    # Skip custom data types, which where already bound under their definition.
    # Skip also data types already bound by RDFLib.
    if iri in set(CUSTOM_TO_PYTHON) | set(OWL_TO_PYTHON_DEFINED_RDFLIB):
        continue
    term.bind(iri, data_type)
# Define what are the Python types compatible with RDF.
ATTRIBUTE_VALUE_TYPES = tuple(x for x in RDF_TO_PYTHON.values())


def rdf_to_python(value: Any, rdf_data_type: URIRef) -> Any:
    """Changes the type of a Python object to match an RDF datatype.

    Converts a Python object into another Python object whose type is
    compatible with the specified data type

    Args:
        value: The Python object to be transformed.
        rdf_data_type: The RDF data type that the resulting object needs to
            be compatible with.

    Returns:
        A Python object compatible with `rdf_data_type`.
    """
    # Converting the literal to a string  or calling Literal twice seems
    #  redundant, but is intentional.
    #
    #  One would expect rdflib to convert the literal to the adequate
    #  datatype by calling 'value.toPython()'. However, this is not always
    #  true. Two cases have to be considered:
    #   - The literal has been created from an arbitrary Python object. Then
    #     calling `toPython` returns the original Python object, no matter
    #     what datatype was specified when creating it.
    #   - The literal has been created from a string. Then calling `toPython`
    #     invokes the RDFLib constructor, spawning a Python object of the
    #     adequate datatype.
    if isinstance(value, Literal):
        # Here we convert the received literal to a string so that RDFLib
        # first calls the lexicalizer on it. In such a way the new literal
        # is created from a string, and the constructor is invoked when
        # calling `toPython`.
        result = Literal(
            str(value), datatype=rdf_data_type, lang=value.language
        ).toPython()
    else:
        # When an arbitrary python object is received instead, first we
        # create a literal with the correct data type, and then perform the
        # same action as before: call the lexicalizer on such literal so
        # that the new one is created from a string.
        result = Literal(value, datatype=rdf_data_type)
        result = Literal(str(result), datatype=rdf_data_type).toPython()
    return result


# --- Various type hints and type hint aliases --- #

# RDF type hints
Pattern = Tuple[Optional[Node], Optional[Node], Optional[Node]]
SimplePattern = Tuple[
    Optional[Union[URIRef, Literal]],
    Optional[Union[URIRef, Literal]],
    Optional[Union[URIRef, Literal]],
]
Triple = Tuple[Node, Node, Node]
SimpleTriple = Tuple[
    Union[URIRef, Literal], Union[URIRef, Literal], Union[URIRef, Literal]
]

# Predicates
OntologyPredicate: TypeAlias = Union[
    "OntologyAnnotation",
    "OntologyAttribute",
    "OntologyRelationship",
]

# Predicate targets
AttributeValue = Union[tuple(value for value in RDF_TO_PYTHON.values())]
AnnotationValue: TypeAlias = Union[
    "OntologyAnnotation",
    "OntologyAttribute",
    "OntologyClass",
    "OntologyIndividual",
    "OntologyRelationship",
    AttributeValue,
    URIRef,
    Literal,
]
RelationshipValue: TypeAlias = "OntologyIndividual"
PredicateValue: TypeAlias = Union[
    "OntologyIndividual", AttributeValue, AnnotationValue
]



"""Classes supporting the definition and use of custom operations in SimPhoNy.

This file contains an `Operations` abstract class that wrapper or package
developers can use to implement specific functionality for certain ontology
classes (e.g. download and upload commands for files, multiplying EMMO
vectors, ...).

Instances of the `OperationsNamespace` class are accessed as the `operations`
property of ontology individuals. The `OperationsNamespace` instances let the
user access the operations defined for each ontology individual. Each
individual has an associated instance of the subclass of `Operations` that the
wrapper or package developer has defined.
"""
from __future__ import annotations

import os
import pkgutil
import sys
from abc import ABC, abstractmethod
from collections.abc import Mapping
from functools import wraps
from importlib import util
from pathlib import Path
from types import ModuleType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from rdflib import URIRef

if sys.version_info < (3, 8):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points

if TYPE_CHECKING:
    from simphony_osp.ontology import OntologyIndividual


__all__ = [
    "Operations",
    "OperationsNamespace",
    "find_operations",
]

_catalog: Dict[URIRef, Dict[str, Tuple[Type, Callable]]] = dict()
"""Holds the operations associated with each ontology class."""

_initialized: List[bool] = [False]
"""True when the installed operations have already been loaded."""


def catalog(func):
    """Initialize the catalog lazily.

    This decorator is meant to decorate functions that write to or access the
    catalog, so that the installed operations can be loaded lazily on the
    first access/write. This is useful to prevent headaches with import cycles.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        if _initialized[0] is False:
            _initialized[0] = True
            _load_operations()
        return func(*args, **kwargs)

    return wrapper


def _load_operations():
    """Finds the installed operations and registers them in the catalog."""
    # Retrieve operations from package entry points.
    package_entry_points = entry_points()
    if sys.version_info >= (3, 10):
        operations = package_entry_points.select(
            group="simphony_osp.ontology.operations"
        )
    else:
        operations = package_entry_points.get(
            "simphony_osp.ontology.operations", tuple()
        )
    del package_entry_points
    operations = {
        entry_point.name: entry_point.load() for entry_point in operations
    }
    for name, operations in operations.items():
        register(operations, operations.iri)
    del operations

    # Retrieve operations from the operation folder in the user's home
    # directory.
    path = (
        os.environ.get("SIMPHONY_OPERATIONS_DIR")
        or Path.home() / ".simphony-osp" / "operations"
    )
    operations = find_operations_in_operations_folder(path)
    for operations in operations:
        register(operations, operations.iri)
    del operations


@catalog
def get(
    item: Union[str, URIRef], default: Optional[Any] = None
) -> Union[Dict[str, Tuple[Type, Callable]], Any]:
    """Get the methods registered for the given identifier.

    Args:
        item: Identifier to get the methods for.
        default: Default value to return when the identifier is not registered.

    Raises:
        KeyError: Identifier not registered and no default provided.
    """
    item = URIRef(item)
    return _catalog.get(item, default)


@catalog
def register(
    class_: Type[Operations], identifier: Union[str, Iterable[str]]
) -> None:
    """Register an `Operations` class in the catalog.

    Args:
        class_: The `Operations` class to register.
        identifier: The identifier (or identifiers) that will be
            registered as associated with the given `Operations` class.

    Raises:
        RuntimeError: Tried to register two methods with the same name for the
            same identifier.
    """
    identifiers = (
        (URIRef(identifier),) if isinstance(identifier, str) else identifier
    )

    methods = class_.__simphony_operations__()

    for identifier in identifiers:
        catalog_entry = _catalog.get(identifier, dict())

        # Raise exception if two methods with the same name are registered
        conflicts = set(catalog_entry) & set(methods)
        if conflicts:
            raise RuntimeError(
                f"Methods {','.join(conflicts)} defined twice for class "
                f"{identifier}."
            )

        # Put the operations on the catalog
        catalog_entry.update(
            {name: (class_, method) for name, method in methods.items()}
        )
        _catalog[identifier] = catalog_entry


class Operations(ABC):
    """Define operations for an ontology class."""

    @property
    @abstractmethod
    def iri(self) -> Union[str, Iterable[str]]:
        """IRI of the ontology class for which operations should be registered.

        It is also possible to define several IRIs at once (by returning an
        iterable).
        """
        pass

    def __init__(self, individual: OntologyIndividual):
        """Initialization of your instance of the operations.

        It is recommended to save the individual that is received as an
        argument to an instance attribute, as the operations to be executed are
        supposed to be related to it.
        """
        self._individual = individual

    @classmethod
    def __simphony_operations__(cls) -> Dict[str, Union[Callable, property]]:
        """Magic method that returns the operations defined on this class."""
        dir_operations = dir(Operations)
        methods = {
            name: getattr(cls, name)
            for name in dir(cls)
            if not (name.startswith("_") or name in dir_operations)
        }
        return methods


class OperationsNamespace(Mapping):
    """Access the operations associated to an ontology individual.

    Instances of the `OperationsNamespace` class are accessed as the
    `operations` property of ontology individuals. The `OperationsNamespace`
    instances let the user access the operations defined for each ontology
    individual. Each individual has an associated instance of the subclass of
    `Operations` that the wrapper or package developer has defined.
    """

    _individual: OntologyIndividual
    _instances: Dict[Type, Operations]

    def __init__(self, individual: OntologyIndividual):
        """Initialize the `OperationsNamespace`."""
        self._instances = dict()
        self._individual = individual

    def __getattr__(self, item: str) -> Any:
        """Get an operation by name using dot notation."""
        try:
            result = self[item]
        except KeyError as e:
            raise AttributeError(str(e)) from e
        return result

    def __setattr__(self, item: str, value: Any) -> None:
        """Set the value of operation's property."""
        if item.startswith("_"):
            super().__setattr__(item, value)
            return

        try:
            self[item] = value
        except KeyError as e:
            raise AttributeError(str(e)) from e

    def __getitem__(self, key: str) -> Union[Callable, Any]:
        """Get an operation by name using brackets."""
        method, instance = self._method_and_instance(key)

        if isinstance(method, property):
            result = getattr(instance, key)
        else:

            @wraps(method)
            def function(*args, **kwargs):
                return method(instance, *args, **kwargs)

            result = function

        return result

    def __setitem__(self, key, value) -> None:
        """Set an operation's property using brackets."""
        method, instance = self._method_and_instance(key)

        if isinstance(method, property):
            setattr(instance, key, value)
        else:
            raise AttributeError(f"operation '{key}' is not writable")

    def __len__(self) -> int:
        """Number of operations available for the individual."""
        return sum(1 for _ in self)

    def __iter__(self) -> Iterator[str]:
        """Iterate over the names of the available operations."""
        yield from {name for name, class_, method in self._all_methods()}

    def _method_and_instance(self, key: str) -> Tuple[Callable, Operations]:
        """Returns the method and operation instance for a given name.

        Searches the catalog for methods with names that match the given key
        and returns both the method and the instance of the `Operations`
        class to which such method belongs.
        """
        results = {
            (class_, method)
            for name, class_, method in self._all_methods()
            if name == key
        }
        if len(results) > 1:
            raise RuntimeError(
                f"More than one operation available under the name {key} for "
                f"individual {self._individual} of classes "
                f"{','.join(str(x) for x in self._individual.classes)} "
                f"available ."
            )
        elif len(results) == 0:
            raise KeyError(
                f"No operation with name {key} available for "
                f"{self._individual} of classes "
                f"{','.join(str(x) for x in self._individual.classes)}."
            )
        class_, method = results.pop()

        if class_ not in self._instances:
            self._instances[class_] = class_(individual=self._individual)
        instance = self._instances[class_]
        return method, instance

    def _all_methods(self) -> Set[Tuple[str, Type, Callable]]:
        """Get a set will all the available operations for the individual."""
        classes = (
            class_.identifier for class_ in self._individual.superclasses
        )
        results = {
            (name, class_, method)
            for identifier in classes
            for name, (class_, method) in get(identifier, dict()).items()
        }
        return results


OPERATIONS = TypeVar("OPERATIONS", bound=Operations)


def find_operations_in_package(
    path: Union[str, Path]
) -> Generator[Type[OPERATIONS]]:
    """Find operations on a Python package.

    Given the path of a Python package (a folder containing an `__init__.py`
    file), this function finds all the operations defined in the package
    and yields them back.

    Args:
        path: location of the Python package to be scanned for operation
            definitions.

    Yields:
        Operation definitions, that is, subclasses of the `Operations` class.
    """
    package_paths = [path]

    def load_submodules_recursively(paths: List[str]) -> Iterator[ModuleType]:
        """Load Python packages and all of their submodules.

        Given the paths of Python packages, this function loads the package as
        well as all the submodules recursively.

        Args:
            paths: Paths of the Python packages to load.

        Yields:
            The loaded Python modules.
        """
        pathlib_paths = [Path(x) for x in paths]
        names = {
            module_info.name
            for module_info in pkgutil.iter_modules(
                str(x.parent.absolute()) for x in pathlib_paths
            )
            if module_info.ispkg
            and module_info.name in {x.name for x in pathlib_paths}
        }
        filter_walk = (
            (loader, module_name, is_pkg)
            for loader, module_name, is_pkg in pkgutil.walk_packages(
                str(x.parent.absolute()) for x in pathlib_paths
            )
            if any(module_name.startswith(name) for name in names)
        )

        for loader, module_name, is_pkg in filter_walk:
            if module_name not in sys.modules:
                spec = loader.find_spec(module_name)
                module = util.module_from_spec(spec)
                spec.loader.exec_module(module)
                sys.modules[module_name] = module
                yield module
            else:
                yield sys.modules[module_name]

    modules = load_submodules_recursively(package_paths)
    modules = {module.__name__ for module in modules}
    classes = (
        x for x in Operations.__subclasses__() if x.__module__ in modules
    )
    return classes


def find_operations_in_operations_folder(
    path: Union[str, Path]
) -> Set[Type[OPERATIONS]]:
    """Find operation definitions in a folder.

    Given a folder path, this function scans the folder for operation
    definitions and returns them.

    Args:
        path: The folder path to be scanned.

    Returns:
        A set of operation definitions, that is, subclasses of the
        `Operations` class.
    """
    package_paths = [path]
    prefix = "simphony_osp.ontology.operations.installed."

    for loader, module_name, is_pkg in pkgutil.walk_packages(
        package_paths, prefix
    ):
        if module_name not in sys.modules:
            spec = loader.find_spec(module_name)
            module = util.module_from_spec(spec)
            spec.loader.exec_module(module)
            sys.modules[module_name] = module
    return {
        operation
        for operation in Operations.__subclasses__()
        if operation.__module__.startswith(prefix)
    }


def find_operations(
    packages: Optional[Union[List[str], str]] = None,
) -> Set[str]:
    """Generates the entry point definitions for operations.

    Scans one or several packages and generates sets of strings that can be
    used in `setup.py` to register SimPhoNy ontology operations.

    This method is meant to ease the work that operation developers have to
    do in their `setup.py` files.

    Args:
        packages: name(s) of the package(s) to scan. When left empty,
            all packages on the working directory are scanned.

    Returns:
        Set of strings that can be used with the
        "simphony_osp.ontology.operations" entry point.

        Example:
            {"File = simphony_osp.ontology.operations.file:File"}
    """
    if isinstance(packages, str):
        packages = [packages]

    path = Path(os.getcwd()).absolute()
    packages = packages or []

    paths = [path / package for package in packages] or [
        path / module_info.name
        for module_info in pkgutil.iter_modules([str(path)])
        if module_info.ispkg
    ]

    operations = {
        op for path in paths for op in find_operations_in_package(path)
    }

    operations = {
        f"{op.__name__} = {op.__module__}:{op.__name__}" for op in operations
    }

    return operations



"""Install SimPhoNy."""

from setuptools import find_packages, setup

# Read description
with open("README.md", "r", encoding="utf8") as readme:
    README_TEXT = readme.read()

NAME = "simphony-osp"
VERSION = "4.0.0rc1"

# main setup configuration class
setup(
    name=NAME,
    version=VERSION,
    description="Framework for creating interfaces between ontologies "
    "and software such as simulation engines, databases and data "
    "repositories.",
    long_description=README_TEXT,
    long_description_content_type="text/markdown",
    url="https://github.com/simphony",
    author="SimPhoNy, EU FP7 Project (Nr. 604005)",
    maintainer="Fraunhofer IWM",
    maintainer_email="simphony@iwm.fraunhofer.de",
    license="BSD-3-Clause",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Operating System :: Unix",
    ],
    keywords="owl, ontology, interoperability, materials-science, osp, "
    "simphony, wrappers, open-simulation-platform,"
    "Fraunhofer IWM",
    download_url="https://pypi.python.org/pypi/simphony-osp",
    project_urls={
        "Tracker": "https://github.com/simphony/simphony-osp/issues",
        "Documentation": "https://simphony.readthedocs.io",
        "Source": "https://github.com/simphony/simphony-osp",
    },
    packages=find_packages(exclude=("examples", "tests")),
    install_requires=[
        "graphviz",
        "numpy",
        "PyYaml",
        "rdflib >= 6.0.2, < 7.0.0",
        "rdflib-sqlalchemy >= 0.5.0",
        "requests",
        "websockets < 11",
        "websockets >= 10; python_version >= '3.10'",
    ],
    python_requires=">=3.7",
    package_data={
        "simphony_osp.ontology.files": [
            "*.yml",
            "*.ttl",
            "*.xml",
            "*.owl",
        ],
    },
    include_package_data=True,
    entry_points={
        "simphony_osp.wrappers": {
            "SQLAlchemy = simphony_osp.interfaces.sqlalchemy:SQLAlchemy",
            "SQLite = simphony_osp.interfaces.sqlite:SQLite",
            "Dataspace = simphony_osp.interfaces.dataspace:Dataspace",
            "Remote = simphony_osp.interfaces.remote:Remote",
        },
        "simphony_osp.ontology.operations": {
            "File = simphony_osp.ontology.operations.file:File",
            "Container = simphony_osp.ontology.operations.container:Container",
        },
        "console_scripts": {
            "pico = simphony_osp.tools.pico:terminal",
            "semantic2dot = simphony_osp.tools.semantic2dot:terminal",
        },
    },
)



"""This file contains keywords that occur in the OWL ontology config files."""

IDENTIFIER_KEY = "identifier"
RDF_FILE_KEY = "ontology_file"
NAMESPACES_KEY = "namespaces"
FILE_FORMAT_KEY = "format"
REQUIREMENTS_KEY = "requirements"

MANDATORY_KEYS = [IDENTIFIER_KEY, RDF_FILE_KEY, NAMESPACES_KEY]

ALL_KEYS = [
    IDENTIFIER_KEY,
    RDF_FILE_KEY,
    NAMESPACES_KEY,
    FILE_FORMAT_KEY,
    REQUIREMENTS_KEY,
]
